/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DatabaseImpl.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.client.xmldb;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.FaultCodes;

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Database;
import org.xmldb.api.base.ErrorCodes;
import org.xmldb.api.base.XMLDBException;

/**
 * DatabaseImpl is the XML:DB driver implementation for Xindice. It is the entry
 * point into the Xindice server but is not intended for direct use by users.
 * Users access it indirectly through the XML:DB DatabaseManager implementation.
 *
 * <p>
 * This API is an implementation of the XML:DB API. More information on this API
 * can be found by looking at
 * <a href="http://xmldb-org.sourceforge.net/xapi/index.html">http://xmldb-org.sourceforge.net/xapi/index.html</a>
 *
 * <p>
 * This implementation simply acts as a proxy for the protocol specific API implementations.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public class DatabaseImpl extends CommonConfigurable implements Database {

    private static final Log log = LogFactory.getLog(DatabaseImpl.class);

    protected Database driver;

    /**
     * These are the URI prefixes this proxy impl understands.
     *
     * XML-RPC is the default.
     */
    public static final String XMLRPC_URI = "xindice://";
    public static final String EMBED_URI = "xindice-embed://";
    public static final String MANAGED_URI = "xindice-managed://";

    /**
     * Name used in the uri for collections associated with this instance.
     */
    public static final String INSTANCE_NAME = "xindice, xindice-embed, xindice-managed";

    /**
     * Name used in the uri for collections associated with this instance.
     */
    private static final String[] INSTANCE_NAMES = {"xindice", "xindice-embed", "xindice-managed"};

    /**
     * The XML:DB API Core Level Conformance of this implementation.
     */
    public static final String CONFORMANCE_LEVEL = "0";


    /**
     * Constructor for the DatabaseImpl object
     */
    public DatabaseImpl() {
        super();
    }

    /**
     * Returns the prefix used in XML:DB to denote URI's that this driver can
     * handle.
     *
     * @return the prefix driver name
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getName() throws XMLDBException {
        return INSTANCE_NAME;
    }

    /**
     * Returns an array of names associated with the Database instance.
     *
     * @return the array of name of the object.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String[] getNames() throws XMLDBException {
        return INSTANCE_NAMES;
    }

    /**
     * Creates a Collection instance using the URI to locate the collection in
     * the Xindice instance. Applications should not call this method directly.
     * Instead they should call {@link org.xmldb.api.DatabaseManager#getCollection(String, String, String)}.
     *
     * <p>
     * The URI format accepted by this method:
     * <code>xindice://[Database Host]/[Database Instance Name]/[Collection Path]</code>
     *
     * <p>
     * Database host is optional.
     *
     * This usually looks something like this:
     * <code>xindice:///db/root/ocs</code> or
     * <code>xindice://some.host.com:8309/db/root/ocs</code>
     *
     * <p>
     * When you pass the URI to DatabaseManager.getCollection(uri) you must
     * prepend an <code>xmldb:</code> to the beginning to make it a valid XML:DB URI.
     * So to normal users of the API URIs will look like this:
     * <code>xmldb:xindice:///db/root/ocs</code>. DatabaseManager will strip the
     * <code>xmldb:</code> before handing the URI to this getCollection implementation.
     *
     * @param uri The URI specifing the location of the collection.
     * @return The Collection value
     * @exception XMLDBException
     */
    public Collection getCollection(String uri, String username, String password) throws XMLDBException {
        createDriver(uri);
        return driver.getCollection(uri, username, password);
    }

    /**
     * Returns the XML:DB API Conformance level for the implementation. This can
     * be used by client programs to determine what functionality is available to
     * them.
     *
     * @return the XML:DB API conformance level for this implementation.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getConformanceLevel() throws XMLDBException {
        return CONFORMANCE_LEVEL;
    }

    /**
     * Determines whether this <code>Database</code> implementation  can handle
     * the URI. It should return true if the Database instance knows how to
     * handle the URI and false otherwise. <p />
     *
     * This method is called by org.xmldb.api.base.DatabaseManager.
     *
     * @param uri the URI to check for.
     * @return true if the URI can be handled, false otherwise.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     * @see #getCollection for a description of the URI's format
     */
    public boolean acceptsURI(String uri) throws XMLDBException {
        if (uri != null && uri.length() > 0 && knownPrefix(uri)) {
            // Ensure the driver is initialised
            createDriver(uri);
            return driver.acceptsURI(uri);
        }

        return false;
    }

    protected void createDriver(String uri) throws XMLDBException {
        // Determine which driver was requested.
        if (driver == null) {
            if (uri == null) {
                throw new XMLDBException(ErrorCodes.INVALID_URI,
                                         FaultCodes.URI_NULL,
                                         "The URI is null");
            }

            if (uri.startsWith(XMLRPC_URI)) {
                /*
                 * The only way that a particular instance of xmlrpc.DatabaseImpl
                 * can be informed of the path to the XML-RPC service in the
                 * web server is by setting a property on the DatabaseImpl object
                 * which is in turn passed to the CollectionImpl object.  Whew!
                 * Since the user never sees the actual xmlrpc.DatabaseImpl object,
                 * this is the only way to make sure that they can set that property.
                 */
                driver = new org.apache.xindice.client.xmldb.xmlrpc.DatabaseImpl(this);
            } else if (uri.startsWith(EMBED_URI)) {
                driver = new org.apache.xindice.client.xmldb.embed.DatabaseImpl(this);
            } else if (uri.startsWith(MANAGED_URI)) {
                driver = new org.apache.xindice.client.xmldb.managed.DatabaseImpl(this);
            } else {
                if (log.isInfoEnabled()) {
                    log.info("The URI '" + uri + "' is not handled by Xindice");
                }
                throw new XMLDBException(ErrorCodes.INVALID_URI,
                                         FaultCodes.URI,
                                         "The URI '" + uri + "' is not handled by Xindice");
            }
        }
    }

    protected boolean knownPrefix(String uri) {
        return (uri.startsWith(XMLRPC_URI) || (uri.startsWith(EMBED_URI)) || uri.startsWith(MANAGED_URI));
    }
}
