/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: CollectionManager.java 571354 2007-08-31 01:49:58Z natalia $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.util.Configurable;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.ConfigurationCallback;
import org.apache.xindice.util.XindiceException;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * CollectionManager is the base class for both Database and Collection.
 *
 * @version $Revision: 571354 $, $Date: 2007-08-30 21:49:58 -0400 (Thu, 30 Aug 2007) $
 */
public class CollectionManager implements Configurable {

    private static final Log log = LogFactory.getLog(CollectionManager.class);

    private static final String COLLECTIONS = "collections";
    private static final String COLLECTION  = "collection";
    private static final String NAME        = "name";

    private static final String[] EMPTY_STRINGS = new String[0];

    private final Map collections = Collections.synchronizedMap(new HashMap());
    private Configuration config;


    protected CollectionManager() {
    }

    /**
     * @param collection a collection to add
     */
    protected void addCollection(Collection collection) {
        this.collections.put(collection.getName(), collection);
    }

    /**
     * Returns number of child collections
     *
     * @return number of collections
     * @throws DBException never
     */
    public final long countCollections() throws DBException {
        return (long) collections.size();
    }

    /**
     * createCollection creates a new Collection object and any associated
     * system resources that the Collection will need.
     *
     * @param path The relative path of the Collection
     * @param cfg The Collection's configuration
     * @return The newly created Collection
     * @throws DBException if operation failed
     */
    public Collection createCollection(String path, Configuration cfg) throws DBException {
        if (path == null || "".equals(path)) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "Null or empty collection name");
        }
        if (cfg == null) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "Error creating collection '" + path + "': null config");
        }
        if (!COLLECTION.equals(cfg.getName())) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "Cannot create collection " + path +
                                  ". Collection configuration top element must be '" + COLLECTION + "'");
        }

        if (path.indexOf("/") != -1) {
            CollectionManager cm = this;
            StringTokenizer st = new StringTokenizer(path, "/");
            while (cm != null && st.hasMoreTokens()) {
                path = st.nextToken().trim();
                if (path.length() == 0) {
                    continue;
                }

                if (!st.hasMoreTokens()) {
                    return cm.createCollection(path, cfg);
                }
                
                cm = (CollectionManager) cm.collections.get(path);
            }

            throw new DBException(FaultCodes.COL_COLLECTION_NOT_FOUND,
                                  "Parent Collection '" + path + "' doesn't exist");
        }

        // Get collection name and do a check
        String n = cfg.getAttribute(NAME);
        if (n == null || n.trim().equals("")) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "No name specified in collection configuration");
        }

        if (!n.equals(path)) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "Name does not match with path name");
        }

        Collection collection = new Collection((Collection) this);
        synchronized (collections) {
            try {
                if (getCollection(n) != null) {
                    throw new DBException(FaultCodes.COL_DUPLICATE_COLLECTION,
                                          "Duplicate Collection '" + n + "'");
                }

                Configuration colConfig = this.config.getChild(COLLECTIONS, true);
                colConfig.add(cfg);

                collection.setConfig(cfg);
                collection.create();
                collections.put(n, collection);
                if (log.isInfoEnabled()) {
                    log.info("Created a new collection named '" + n + "'");
                }
            } catch (DBException e) {
                // Do not wrap DBException
                throw e;
            } catch (Exception e) {
                throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                      "Error Creating Collection '" + path + "'", e);
            }

            return collection;
        }
    }

    public boolean close() throws DBException {
        synchronized (collections) {
            for (Iterator i = collections.values().iterator(); i.hasNext(); ) {
                Collection collection = (Collection) i.next();
                try {
                    collection.close();
                } catch (DBException e) {
                    if (log.isWarnEnabled()) {
                        log.warn("ignored exception", e);
                    }
                }
            }
        }

        return true;
    }

    /**
     * dropCollection physically removes the specified Collection and any
     * associated system resources that the Collection uses.
     *
     * @param collection The Collection to drop
     * @return Whether or not the Collection was dropped
     * @throws DBException if operation failed
     */
    public boolean dropCollection(Collection collection) throws DBException {
        if (collection == null) {
            throw new DBException(FaultCodes.COL_COLLECTION_NOT_FOUND,
                                  "Collection Value Null");
        }

        Collection cm = collection.getParentCollection();
        if (cm == null) {
            throw new DBException(FaultCodes.DBE_CANNOT_DROP,
                                  "You Cannot Drop The Database");
        }

        if (cm != this) {
            return cm.dropCollection(collection);
        }

        final String name = collection.getName();

        synchronized (collections) {
            boolean dropped = collection.drop();
            if (dropped) {
                collections.remove(name);
                Configuration colConfig = config.getChild(COLLECTIONS);
                colConfig.processChildren(COLLECTION, new ConfigurationCallback() {
                    public void process(Configuration cfg) {
                        try {
                            if (cfg.getAttribute(NAME).equals(name)) {
                                cfg.delete();
                            }
                        } catch (Exception e) {
                            if (log.isWarnEnabled()) {
                                log.warn("ignored exception", e);
                            }
                        }
                    }
                });
            }

            return dropped;
        }
    }

    /**
     * getCollection retrieves a Collection by name.
     *
     * @param path The Collection path
     * @return The Collection (or null)
     * @throws DBException if operation failed
     */
    public Collection getCollection(String path) throws DBException {
        if (path == null) {
            return null;
        }

        if (path.indexOf("/") != -1) {
            CollectionManager cm = this;
            StringTokenizer st = new StringTokenizer(path, "/");
            while (cm != null && st.hasMoreTokens()) {
                path = st.nextToken().trim();
                if (path.length() == 0) {
                    continue;
                }

                cm = (CollectionManager) cm.collections.get(path);
            }

            return (Collection) cm;
        }
        
        return (Collection) collections.get(path);
    }

    /**
     * @return The collections map
     */
    protected Map getCollections() {
        return Collections.unmodifiableMap(this.collections);
    }

    /**
     * @see org.apache.xindice.util.Configurable#getConfig()
     */
    public Configuration getConfig() {
        return config;
    }

    /**
     * listCollections retrieves a list of Collections as an array of
     * Strings.
     *
     * @return The Collection list
     * @throws DBException never
     */
    public final String[] listCollections() throws DBException {
        synchronized (collections) {
            return (String[]) collections.keySet().toArray(EMPTY_STRINGS);
        }
    }

    /**
     * @see org.apache.xindice.util.Configurable#setConfig(org.apache.xindice.util.Configuration)
     */
    public void setConfig(Configuration config) throws XindiceException {
        this.config = config;

        Configuration colConfig = config.getChild(COLLECTIONS);
        if (colConfig != null) {
            colConfig.processChildren(COLLECTION, new ConfigurationCallback() {
                public void process(Configuration cfg) throws XindiceException {
                    // check for an existing Collection by name and, if found, skip creating a new Collection
                    // else, create a new child Collection, configure it an add it
                    // if the Collection already exists in our collections list,
                    // creating a new one will cause the old one to be discarded
                    // while holding open file handles, etc.
                    Collection col = (Collection) collections.get(cfg.getAttribute(NAME));
                    if (col == null) {
                        col = new Collection((Collection) CollectionManager.this);
                        col.setConfig(cfg);
                        collections.put(col.getName(), col);
                    }
                    // else, assume col is configured elsewhere...
                    // I'm not sure this should be happening, but it does
                    // it is not safe to call col.setConfig again since it does a bunch
                    // of stuff (like creating a new Filer and overwriting the existing one, etc.)
                }
            });
        }
    }
}
