/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: IndexPattern.java 568786 2007-08-23 00:40:57Z vgritsenko $
 */

package org.apache.xindice.core.indexer;

import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.filer.Streamable;
import org.apache.xindice.util.ReadOnlyException;
import org.apache.xindice.xml.NamespaceMap;
import org.apache.xindice.xml.SymbolTable;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.StringTokenizer;

/**
 * IndexPattern is the internal representation of a pattern for index
 * matching purposes.
 *
 * @version $Revision: 568786 $, $Date: 2007-08-22 20:40:57 -0400 (Wed, 22 Aug 2007) $
 */
public final class IndexPattern implements Streamable {

    public static final int PATTERN_NONE     = -1;
    public static final int PATTERN_WILDCARD = -2;
    public static final int PATTERN_NAME     = -3;

    public static final int SCORE_NONE     = 0;
    public static final int SCORE_WILDCARD = 1;
    public static final int SCORE_NAME     = 2;
    public static final int SCORE_NATURAL  = 3;


    private SymbolTable symbols;
    private String elemName;
    private short elemID = PATTERN_NONE;
    private String attrName;
    private short attrID = PATTERN_NONE;


    public IndexPattern(SymbolTable symbols, String pattern, NamespaceMap nsMap) throws IndexerException {
        this.symbols = symbols;
        StringTokenizer st = new StringTokenizer(pattern.trim(), "@");

        try {
            elemName = st.nextToken();
            if (elemName.equals("*")) {
                elemID = PATTERN_WILDCARD;
            } else {
                elemID = symbols.getNormalizedSymbol(elemName, nsMap, true);
            }

            if (st.hasMoreTokens()) {
                attrName = st.nextToken();
                if (attrName.equals("*")) {
                    attrID = PATTERN_WILDCARD;
                } else {
                    attrID = elemID == PATTERN_WILDCARD ? PATTERN_NAME
                            : symbols.getNormalizedSymbol(attrName, nsMap, true);
                }
            }
        } catch (ReadOnlyException e) {
            throw new IndexerException(FaultCodes.IDX_NOT_SUPPORTED,
                                       "Pattern '" + pattern + "' is not allowed in this collection.", e);
        }
    }

    public IndexPattern(SymbolTable symbols, short elemID) {
        this.symbols = symbols;
        this.elemID = elemID;
        this.elemName = symbols.getName(elemID);
    }

    public IndexPattern(SymbolTable symbols, short elemID, short attrID) {
        this.symbols = symbols;
        this.elemID = elemID;
		this.elemName = symbols.getName(elemID);
        this.attrID = attrID;
		this.attrName = symbols.getName(attrID);
    }

    public IndexPattern(SymbolTable symbols, short elemID, String attrName) {
        this.symbols = symbols;
        this.elemID = elemID;
        this.attrID = PATTERN_NAME;
        this.attrName = attrName;
    }

    /**
     * getMatchLevel compares this IndexPattern to p.
     *
     * It returns the strength of
     * the match betwen the two patterns from 0 to 15.  (0 being no match at
     * all, 15 being perfect).  This method should be called by XPath and
     * SAX patterns to be matched against Indexer patterns.
     *
     * @param p The pattern to compare
     * @return The resulting IndexPattern strength
     */
    public int getMatchLevel(IndexPattern p) {
        int result = 0;

        switch (p.elemID) {
            case PATTERN_WILDCARD:
                result = (SCORE_WILDCARD << 2);
                break;
            case PATTERN_NAME:
                if (p.elemName.equals(elemName)) {
                    result = (SCORE_NAME << 2);
                }
                break;
            default:
                if (elemID == p.elemID) {
                    result = (SCORE_NATURAL << 2);
                }
        }

        if (result != 0) {
            switch (p.attrID) {
                case PATTERN_WILDCARD:
                    return result + SCORE_WILDCARD;
                case PATTERN_NAME:
                    if (p.attrName.equals(attrName)) {
                        return result + SCORE_NAME;
                    }
                    break;
                default:
                    if (attrID == p.attrID) {
                        return result + SCORE_NATURAL;
                    }
            }
        }

        return 0;
    }

    /**
     * getElementID returns the Element Symbol ID for this pattern if
     * there is one, otherwise it returns a negative value.
     *
     * @return The Element Symbol ID
     */
    public short getElementID() {
        return elemID;
    }

    /**
     * getAttributeID returns the Attribute Symbol ID for this pattern
     * if there is one, otherwise it returns a negative value.
     *
     * @return The Attribute Symbol ID
     */
    public short getAttributeID() {
        return attrID;
    }

    /**
     * getElementName returns the Element Name for this pattern if
     * there is one, otherwise it returns null.
     *
     * @return The Element Name
     */
    public String getElementName() {
        return elemName;
    }

    /**
     * getAttributeName returns the Attribute Name for this pattern if
     * there is one, otherwise it returns null.
     *
     * @return The Attribute Name
     */
    public String getAttributeName() {
        return attrName;
    }

    public void read(DataInput dis) throws IOException {
        elemID = dis.readShort();
        if (elemID == PATTERN_NAME) {
            short len = dis.readShort();
            byte[] name = new byte[len];
            dis.readFully(name);
            elemName = new String(name, "utf-8");
        }
        attrID = dis.readShort();
        if (attrID == PATTERN_NAME) {
            short len = dis.readShort();
            byte[] name = new byte[len];
            dis.readFully(name);
            attrName = new String(name, "utf-8");
        }
    }

    public void write(DataOutput dos) throws IOException {
        dos.writeShort(elemID);
        if (elemID == PATTERN_NAME) {
            byte[] name = elemName.getBytes("utf-8");
            dos.writeShort(name.length);
            dos.write(name);
        }
        dos.writeShort(attrID);
        if (attrID == PATTERN_NAME) {
            byte[] name = attrName.getBytes("utf-8");
            dos.writeShort(name.length);
            dos.write(name);
        }
    }

    public int hashCode() {
        return (elemID << 16) + attrID;
    }

    public boolean equals(Object obj) {
        if (obj instanceof IndexPattern) {
            IndexPattern p = (IndexPattern) obj;
            boolean eq = elemID == p.elemID && attrID == p.attrID;
            if (eq && (elemName != null || p.elemName != null)) {
                eq = elemName != null && p.elemName != null && elemName.equals(p.elemName);
            }
            if (eq && (attrName != null || p.attrName != null)) {
                eq = attrName != null && p.attrName != null && attrName.equals(p.attrName);
            }
            return eq;
        } else {
            return false;
        }
    }
}
