/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: SpecialQueryParser.java 586647 2007-10-20 00:32:43Z natalia $
 */

package org.apache.xindice.core.query.ftsearch;

import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.queryParser.CharStream;
import org.apache.lucene.queryParser.QueryParserTokenManager;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.BooleanClause;

import java.util.Vector;

/**
 * Specialized parser for text queries that ignores query clauses that have
 * 'prohibited' modifier. This is only used for XPath full text search
 * extention function to search existing text index, if any.
 *
 * @version $Revision: 586647 $, $Date: 2007-10-19 20:32:43 -0400 (Fri, 19 Oct 2007) $
 */
public class SpecialQueryParser extends QueryParser {
    private static final int CONJ_AND    = 1;
    private static final int CONJ_OR     = 2;

    private static final int MOD_NOT     = 10;
    private static final int MOD_REQ     = 11;

    public SpecialQueryParser(String f, Analyzer a) {
        super(f, a);
    }

    public SpecialQueryParser(CharStream stream) {
        super(stream);
    }

    public SpecialQueryParser(QueryParserTokenManager tm) {
        super(tm);
    }

    /**
     * This method is slightly modified copy of superclass method, where it ignores
     * boolean clauses that have 'prohibited' modifier.
     * @see QueryParser#addClause(java.util.Vector, int, int, org.apache.lucene.search.Query)
     */
    protected void addClause(Vector clauses, int conj, int mods, Query q) {

        // If this term is introduced by AND, make the preceding term required,
        // unless it's already prohibited
        if (clauses.size() > 0 && conj == CONJ_AND) {
            BooleanClause c = (BooleanClause) clauses.elementAt(clauses.size()-1);
            if (!c.isProhibited())
                c.setOccur(BooleanClause.Occur.MUST);
        }

        if (clauses.size() > 0 && getDefaultOperator() == AND_OPERATOR && conj == CONJ_OR) {
            // If this term is introduced by OR, make the preceding term optional,
            // unless it's prohibited (that means we leave -a OR b but +a OR b-->a OR b)
            // notice if the input is a OR b, first term is parsed as required; without
            // this modification a OR b would parsed as +a OR b
            BooleanClause c = (BooleanClause) clauses.elementAt(clauses.size()-1);
            if (!c.isProhibited())
                c.setOccur(BooleanClause.Occur.SHOULD);
        }

        // We might have been passed a null query; the term might have been
        // filtered away by the analyzer.
        if (q == null)
            return;

        boolean required, prohibited;
        if (getDefaultOperator() == OR_OPERATOR) {
            // We set REQUIRED if we're introduced by AND or +; PROHIBITED if
            // introduced by NOT or -; make sure not to set both.
            prohibited = (mods == MOD_NOT);
            required = (mods == MOD_REQ);
            if (conj == CONJ_AND && !prohibited) {
                required = true;
            }
        } else {
            // We set PROHIBITED if we're introduced by NOT or -; We set REQUIRED
            // if not PROHIBITED and not introduced by OR
            prohibited = (mods == MOD_NOT);
            required = (!prohibited && conj != CONJ_OR);
        }
        if (required && !prohibited) {
            clauses.addElement(new BooleanClause(q, BooleanClause.Occur.MUST));
        } else if (!required && !prohibited) {
            clauses.addElement(new BooleanClause(q, BooleanClause.Occur.SHOULD));
        } else if (required && prohibited) {
            throw new RuntimeException("Clause cannot be both required and prohibited");
        }
    }
}
