/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: AddCollection.java 578602 2007-09-23 20:33:31Z natalia $
 */

package org.apache.xindice.tools.command;

import org.apache.xindice.client.xmldb.services.CollectionManager;
import org.apache.xindice.tools.XMLTools;
import org.apache.xindice.xml.dom.DocumentImpl;
import org.apache.xindice.xml.dom.DOMParser;
import org.apache.xindice.xml.TextWriter;
import org.apache.xindice.util.XindiceException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Attr;
import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;

import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;

/**
 * AddCollection.java is designed to let the user create
 * new Collections and Nested Collections within the Database.
 *
 * @version $Revision: 578602 $, $Date: 2007-09-23 16:33:31 -0400 (Sun, 23 Sep 2007) $
 */
public class AddCollection extends Command {

    /**
     * Creates collection
     */
    public boolean execute(XMLTools.Config table) throws Exception {

        if (table.getString(XMLTools.COLLECTION) == null) {
            System.out.println("ERROR : Collection and switch required");
            return false;
        }

        if ("".equals(table.getString(XMLTools.FILE_PATH)) && table.getString(XMLTools.NAME_OF) == null) {
            System.out.println("ERROR : New Collection name required or File path required");
            return false;
        }

        Document config;
        if (!"".equals(table.getString(XMLTools.FILE_PATH))) {
            // configuration was passed in a file
            config = readConfig(table);
        } else {
            // build configuration from parameters
            config = buildConfig(table);
        }


        Collection col = null;
        Collection tempcol = null;
        try {
            // Get a Collection reference to pass on to individual commands
            String colstring = normalizeCollectionURI(table.getString(XMLTools.COLLECTION),
                                                      table.getBoolean(XMLTools.LOCAL));

            col = DatabaseManager.getCollection(colstring);
            if (col == null) {
                System.out.println("ERROR : Collection not found!");
                return false;
            }

            // Create an instance of the collection manager and create the collection
            CollectionManager colman = (CollectionManager) col.getService("CollectionManager", XMLDBAPIVERSION);

            String colPath;
            if (!"".equals(table.getString(XMLTools.FILE_PATH))) {
                colPath = config.getDocumentElement().getAttributeNode("name").getNodeValue();
            } else {
                colPath = table.getString(XMLTools.NAME_OF);
            }

            tempcol = colman.createCollection(colPath, config);

            System.out.println("Created : " + table.getString(XMLTools.COLLECTION) + "/" + colPath);
        } finally {
            // Release the collection objects
            if (col != null) {
                col.close();
            }
            if (tempcol != null) {
                tempcol.close();
            }
        }

        return true;
    }

    private Document readConfig(XMLTools.Config table) throws Exception {
        InputStream fis = null;
        try {
            File file = new File(table.getString(XMLTools.FILE_PATH));
            fis = new FileInputStream(file);

            // check if top element has necessary attribute 'name'
            Document doc = DOMParser.toDocument(fis);
            Attr attr = doc.getDocumentElement().getAttributeNode("name");
            if (attr == null || "".equals(attr.getNodeValue())) {
                throw new Exception("Collection configuration top element must have attribute \"name\"");
            }

            return doc;
        } catch (XindiceException e) {
            throw new Exception("Indexer configuration could not be parsed", e);
        } finally {
            if (fis != null) {
                fis.close();
            }
        }
    }

    private Document buildConfig(XMLTools.Config table) throws Exception {
        String colPath = table.getString(XMLTools.NAME_OF);

        String colName;
        // get the part after last /, any collection before / must already exist
        int idx = colPath.lastIndexOf("/");
        if (idx != -1) {
            colName = colPath.substring(idx + 1);
        } else {
            colName = colPath;
        }

        if (colName.equals("")) {
            throw new Exception("Cannot create a NULL collection");
        }

        Document doc = new DocumentImpl();

        Element colEle = doc.createElement("collection");
        colEle.setAttribute("name", colName);
        // FIXME Make this configurable
        colEle.setAttribute("compressed", "true");
        colEle.setAttribute("inline-metadata", "true");

        doc.appendChild(colEle);

        Element filEle = doc.createElement("filer");
        String filerClass = "org.apache.xindice.core.filer.BTreeFiler";
        // see if user specified filer type
        if (table.getString(XMLTools.FILER) != null) {
            String filer = table.getString(XMLTools.FILER);
            if ("HashFiler".equals(filer)) {
                filerClass = "org.apache.xindice.core.filer.HashFiler";
            } else if (!"BTreeFiler".equals(filer)) {
                throw new Exception("Unknown filer: " + filer);
            }
        }

        filEle.setAttribute("class", filerClass);
        if (table.getString(XMLTools.PAGE_SIZE) != null) {
            filEle.setAttribute(XMLTools.PAGE_SIZE, table.getString(XMLTools.PAGE_SIZE));
        }
        if (table.getString(XMLTools.MAX_KEY_SIZE) != null) {
            filEle.setAttribute(XMLTools.MAX_KEY_SIZE, table.getString(XMLTools.MAX_KEY_SIZE));
        }
        if (table.getString(XMLTools.PAGE_COUNT) != null) {
            filEle.setAttribute(XMLTools.PAGE_COUNT, table.getString(XMLTools.PAGE_COUNT));
        }

        colEle.appendChild(filEle);

        // If in verbose mode, show....
        if (table.getBoolean(XMLTools.VERBOSE)) {
            String colstr = TextWriter.toString(doc);
            System.out.println("Collection node element = ");
            System.out.println("\t" + colstr + "\n");
        }

        return doc;
    }
    public void usage() {
        System.out.println("Format: xindice ac -c <context> [-l [-d <path>]] [-v] [parameters...]");
        System.out.println();
        System.out.println("Creates new collections and nested collections");
        System.out.println();
        System.out.println("Command-specific switches:");
        System.out.println("    -n|--nameOf <name>");
        System.out.println("                 Name for the collection to be created, must be present");
        System.out.println("                 unless collection configuration is specified");
        System.out.println("    -f|--filepath <file>");
        System.out.println("                 Name of the file that holds collection configuration. If");
        System.out.println("                 specified, the rest of command-specific parameters is");
        System.out.println("                 ignored");
        System.out.println("    --pagesize <number>");
        System.out.println("                 Page size for file pages (default: 4096)");
        System.out.println("    --maxkeysize <number>");
        System.out.println("                 The maximum size for file keys (default: 0=none)");
        System.out.println("    --pagecount <nuymber>");
        System.out.println("                 Number of pages in the primary storage (default: 1024)");
        System.out.println("    --filer <name>");
        System.out.println("                 Collection filer, can be HashFiler or BTreeFiler (default:");
        System.out.println("                 BTreeFiler)");
        System.out.println();
    }
}