/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: ImportTree.java 578602 2007-09-23 20:33:31Z natalia $
 */

package org.apache.xindice.tools.command;

import org.apache.xindice.tools.XMLTools;

import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;

import java.io.File;
import java.io.FileFilter;

/**
 * ImportTree.java is designed to take a directory/file path and create a Collection
 * from it's name. The files within the directory are converted into Documents
 * and then stored into the newly created Collection with their filename as a document key
 *
 * NOTE:  If the collection passed in with the -c option already exists it will be overwritten!
 *
 * @version $Revision: 578602 $, $Date: 2007-09-23 16:33:31 -0400 (Sun, 23 Sep 2007) $
 */
public class ImportTree extends Command {

    Command addDocument = null;
    Command addCollection = null;

    public boolean execute(XMLTools.Config table) throws Exception {

        // Verify that the correct command line parameters were passed in
        if (table.getString(XMLTools.FILE_PATH) == null) {
            System.out.println("ERROR : Directory name and switch required");
            return false;
        }

        if (table.getString(XMLTools.COLLECTION) == null) {
            System.out.println("ERROR : Collection name and switch required");
            return false;
        }

        Collection col = null;
        try {
            // Setup the creation commands we'll use to add documents and create
            // collections.
            addDocument =
                    (Command) Class.forName("org.apache.xindice.tools.command.AddDocument").newInstance();

            addCollection =
                    (Command) Class.forName("org.apache.xindice.tools.command.AddCollection").newInstance();

            String startcollection = table.getString(XMLTools.COLLECTION);

            // Make sure the root collection exists
            String colstring = normalizeCollectionURI(startcollection, table.getBoolean(XMLTools.LOCAL));
            if ((col = DatabaseManager.getCollection(colstring)) == null) {
                System.out.println("ERROR : Collection not found!");
                return false;
            }

            File startdir = new File(table.getString(XMLTools.FILE_PATH));

            // Make call to process, this is called recursively!
            process(startdir, colstring, table);

        } finally {
            // Be sure to close collection objects
            if (col != null) {
                col.close();
            }
        }

        return true;
    }

    // Recursive function to create collections/documents
    void process(File directory, String baseCollection, XMLTools.Config table) {
        try {
            // The file extension to use for reading in files
            final String ext = table.getString(XMLTools.EXTENSION);

            // Make sure we received a directory so we don't create a collection
            // named after a file.
            if (!directory.isDirectory()) {
                System.out.println("ERROR: The specified import path is not a directory " +
                                   directory.getPath());
                return;
            }

            // Create the collection for this directory.
            baseCollection = createCollection(baseCollection, directory.getName());

            // Get a directory listing and loop over
            File[] subdirs = directory.listFiles(new ExtensionFileFilter(ext));

            for (int i = 0; i < subdirs.length; i++) {
                // Is this a directory?  If so, call ourselves recursively
                if (subdirs[i].isDirectory()) {
                    process(subdirs[i], baseCollection, table);
                } else {
                    // Otherwise we have a file so import it.
                    importFile(baseCollection, subdirs[i].getAbsolutePath(), subdirs[i].getName());
                }
            }
        } catch (Exception e) {
            System.out.println("ERROR : " + e.getMessage());
            if (table.getBoolean(XMLTools.VERBOSE)) {
                e.printStackTrace(System.err);
            }
        }
    }

    protected void importFile(String baseCollection, String path, String name)
            throws Exception {
        XMLTools.Config table = new XMLTools.Config();
        // Use functionality from AddDocument to add this document
        table.setString(XMLTools.COLLECTION, baseCollection);
        table.setString(XMLTools.FILE_PATH, path);
        table.setString(XMLTools.NAME_OF, name);

        addDocument.execute(table);
    }

    protected String createCollection(String baseCollection, String newCollection)
            throws Exception {
        if (!newCollection.equals(".")) {
            XMLTools.Config table = new XMLTools.Config();
            table.setString(XMLTools.COLLECTION, baseCollection);
            table.setString(XMLTools.NAME_OF, newCollection);

            addCollection.execute(table);
        } else {
            newCollection = "";
        }

        return baseCollection + "/" + newCollection;
    }

    /**
     * Allow us to filter directory listings over certain file extensions
     */
    private static class ExtensionFileFilter implements FileFilter {

        private String extension = "";

        /**
         * Constructs an <code>MyFileFiler</code> instance with a specific
         * extension.
         *
         * @param extension The file's extension
         */
        public ExtensionFileFilter(String extension) {
            this.extension = extension;
        }

        /**
         * Tests whether or not the specified abstract pathname should be
         * included in a pathname list.
         *
         * @param  pathname  The abstract pathname to be tested
         * @return  <code>true</code> if and only if <code>pathname</code>
         *          should be included
         */
        public boolean accept(File pathname) {
            if (this.extension.length() > 0) {
                return pathname.getName().endsWith("." + this.extension) || pathname.isDirectory();
            } else {
                return true;
            }
        }
    }

    public void usage() {
        System.out.println("Format: xindice import -c <context> [-l [-d <path>]] [-v] [parameters...]");
        System.out.println();
        System.out.println("Takes a directory path and creates a collection from it's name. The files");
        System.out.println("within the directory are converted into documents and then stored into the");
        System.out.println("newly created collection with their filename as a document key.");
        System.out.println();
        System.out.println("Command-specific switches:");
        System.out.println("    -f|--filepath <file>");
        System.out.println("                 Directory tree");
        System.out.println("    -e|--extension <extention>");
        System.out.println("                 File extension for documents to use in the filter");
        System.out.println();
    }
}
