/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: RetrieveDocument.java 578602 2007-09-23 20:33:31Z natalia $
 */

package org.apache.xindice.tools.command;

import org.apache.xindice.tools.XMLTools;

import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;
import org.xmldb.api.modules.XMLResource;

import java.io.File;
import java.io.FileOutputStream;

/**
 * RetrieveDocument.java is designed to let the user recall or
 * retrieve a Document from a specific Collection and save it to a new location.
 *
 * @version $Revision: 578602 $, $Date: 2007-09-23 16:33:31 -0400 (Sun, 23 Sep 2007) $
 */
public class RetrieveDocument extends Command {

    public boolean execute(XMLTools.Config table) throws Exception {

        if (table.getString(XMLTools.COLLECTION) == null) {
            System.out.println("ERROR : Collection name and switch required");
            return false;
        }

        if (table.getString(XMLTools.NAME_OF) == null) {
            System.out.println("ERROR : Document Key and switch required");
            return false;
        }

        Collection col = null;
        try {

            // Create a collection instance
            String colstring = normalizeCollectionURI(table.getString(XMLTools.COLLECTION),
                                                      table.getBoolean(XMLTools.LOCAL));
            col = DatabaseManager.getCollection(colstring);
            // Make sure that the collection was found
            if (col == null) {
                System.out.println("ERROR : Collection not found!");
                return false;
            }

            String docname = table.getString(XMLTools.NAME_OF);
            XMLResource resource = (XMLResource) col.getResource(docname);
            // Verify that we were able to find the document
            if (resource == null) {
                System.out.println("ERROR : Document not found!");
                return false;
            }

            String documentstr = (String) resource.getContent();
            String path = table.getString(XMLTools.FILE_PATH);
            if (documentstr != null && path != null && !"".equals(path)) {

                FileOutputStream out = null;
                try {
                    File file = new File(table.getString(XMLTools.FILE_PATH));
                    // Create the directory structure if necessary
                    if (file.getParentFile() != null) {
                        file.getParentFile().mkdirs();
                    }

                    out = new FileOutputStream(file);

                    System.out.println("Writing...");

                    out.write(documentstr.getBytes("utf-8"));
                    out.close();

                    System.out.println("Wrote file " + table.getString(XMLTools.FILE_PATH));
                } finally {
                    if (out != null) {
                        out.close();
                    }
                }
            } else {
                // Send to stdout
                System.out.println(documentstr);
            }
        } finally {
            // Release the collection object
            if (col != null) {
                col.close();
            }
        }

        return true;
    }

    public void usage() {
        System.out.println("Format: xindice rd  -c <context> [-l [-d <path>]] [-v] [parameters...]");
        System.out.println();
        System.out.println("Retrieves a document from a specific collection");
        System.out.println();
        System.out.println("Command-specific switches:");
        System.out.println("    -f|--filepath <file>");
        System.out.println("                 File path where document will be stored, if omitted");
        System.out.println("                 document will printed to standard output");
        System.out.println("    -n|--nameOf <name>");
        System.out.println("                 Document key");
        System.out.println();
        System.out.println("Examples:");
        System.out.println("    xindice rd -c /db/test/ocs -f a:\\file.xml -n file.xml");
        System.out.println();
    }
}

