/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: LockManager.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.util;

import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * LockManager manages resource locks.  A resource is any shared object that
 * can be represented as a long.  The LockManager should be constructed using
 * the maximum number of concurrent clients a specific resource might have.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public final class LockManager {

    private static final Log log = LogFactory.getLog(LockManager.class);

    private int maxLocks = 0;
    private Map locks = new HashMap(); // Long to LockInfo

    public LockManager(int maxLocks) {
        this.maxLocks = maxLocks;
    }

    /**
     * acquireSharedLock acquires a shared lock on a resource.  Shared locks
     * are typically used for reads.
     */
    public void acquireSharedLock(long id) {
        acquireSharedLock(new Long(id));
    }

    /**
     * acquireSharedLock acquires a shared lock on a resource.  Shared locks
     * are typically used for reads.
     */
    public void acquireSharedLock(Object key) {
        LockInfo info;
        synchronized (this) {
            info = (LockInfo) locks.get(key);
            if (info == null) {
                info = new LockInfo();
                locks.put(key, info);
            }
        }
        synchronized (info) {
            while (info.count >= maxLocks) {
                try {
                    info.wait();
                } catch (Exception e) {
                    if (log.isWarnEnabled()) {
                        log.warn("ignored exception", e);
                    }
                }
            }
            info.count++;
        }
    }

    /**
     * releaseSharedLock releases a shared lock on a resource.  Shared locks
     * are typically used for reads.
     */
    public void releaseSharedLock(long id) {
        releaseSharedLock(new Long(id));
    }

    /**
     * releaseSharedLock releases a shared lock on a resource.  Shared locks
     * are typically used for reads.
     */
    public void releaseSharedLock(Object key) {
        LockInfo info;
        synchronized (this) {
            info = (LockInfo) locks.get(key);
        }
        if (info == null) {
            return;
        }
        
        synchronized (info) {
            info.count--;
            info.notify();
            if (info.count == 0) {
                synchronized (this) {
                    locks.remove(key);
                }
            }
        }
    }

    /**
     * acquireExclusiveLock acquires an exclusive lock on a resource.  Exclusive
     * locks are typically used for writes.
     */
    public void acquireExclusiveLock(long id) {
        acquireExclusiveLock(new Long(id));
    }

    /**
     * acquireExclusiveLock acquires an exclusive lock on a resource.  Exclusive
     * locks are typically used for writes.
     */
    public void acquireExclusiveLock(Object key) {
        LockInfo info;
        synchronized (this) {
            info = (LockInfo) locks.get(key);
            if (info == null) {
                info = new LockInfo();
                locks.put(key, info);
            }
        }
        synchronized (info) {
            int total = 0;
            int stolen = 0;
            while (total < maxLocks && info.count >= maxLocks) {
                try {
                    info.wait();
                } catch (Exception e) {
                    if (log.isWarnEnabled()) {
                        log.warn("ignored exception", e);
                    }
                }
                stolen = maxLocks - info.count;
                total += stolen;
                info.count += stolen;
            }
        }
    }

    /**
     * releaseExclusiveLock releases an exclusive lock on a resource.  Exclusive
     * locks are typically used for writes.
     */
    public void releaseExclusiveLock(long id) {
        releaseExclusiveLock(new Long(id));
    }

    /**
     * releaseExclusiveLock releases an exclusive lock on a resource.  Exclusive
     * locks are typically used for writes.
     */
    public void releaseExclusiveLock(Object key) {
        LockInfo info;
        synchronized (this) {
            info = (LockInfo) locks.get(key);
        }
        if (info == null) {
            return;
        }
        
        synchronized (info) {
            info.count = 0;
            info.notify();
            if (info.count == 0) {
                synchronized (this) {
                    locks.remove(key);
                }
            }
        }
    }

    /**
     * LockInfo is just an int wrapper.
     */
    private class LockInfo {
        public int count = 0;
    }
}
