/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: XPathSearchCollectionViewer.java 541515 2007-05-25 02:45:06Z vgritsenko $
 */

package org.apache.xindice.webadmin.viewer.components;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Node;

import org.apache.xindice.xml.NamespaceMap;
import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.data.NodeSet;
import org.apache.xindice.webadmin.viewer.*;
import org.apache.xindice.webadmin.viewer.HtmlViewerComponent;
import org.apache.xindice.xml.Xml2HtmlWriter;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Xindice Html Viewer for collection xpath searches.
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @version $Revision: 541515 $, $Date: 2007-05-24 22:45:06 -0400 (Thu, 24 May 2007) $
 */
public class XPathSearchCollectionViewer extends HtmlViewerComponent implements HtmlCollectionViewer {

    private static final Log log = LogFactory.getLog(XPathSearchCollectionViewer.class);

    public void execute(HttpServletRequest req, HttpServletResponse res, Collection col)
            throws ServletException, IOException {

        String query = req.getParameter("query");
        if (query == null) {
            query = "";
        }

        String path = col.getCanonicalName();
        String title = "XPATH Query for " + path;
        ServletOutputStream output = startViewer(title, path, res);

        // get namespaces
        String[] ns = {"", ""};
        int nsSize = 1;
        String rawNsSize = req.getParameter("nssize");
        if (rawNsSize != null) {
            nsSize = Integer.parseInt(rawNsSize);
            ns = new String[nsSize * 2];
            int index = 0;
            for (int i = 1; i <= nsSize; i++) {
                String prefix = req.getParameter("prefix" + i);
                String uri = req.getParameter("uri" + i);
                if (prefix != null) {
                    ns[index] = prefix;
                } else {
                    ns[index] = "";
                }
                index++;
                if (uri != null) {
                    ns[index] = uri;
                } else {
                    ns[index] = "";
                }
                index++;
            }
        }

        printQueryForm(output, query, path, ns);
        output.print("<br />\n");
        output.flush();

        if (query.length() > 0) { // print result
            NamespaceMap nsMap = new NamespaceMap();
            int index = 0;
            for (int i = 0; i < nsSize; i++) {
                nsMap.setNamespace(ns[index++], ns[index++]);
            }
            try {
                NodeSet result = col.queryCollection("XPath", query, nsMap);
                printQuery(result, output, query, path);
            } catch (DBException e) {
                log.error(e);
                printStartTable("Error while XPath Query", output);
                printStartSingleTableBox(output);
                output.print("<b>Error while XPath Query:</b><br />");
                output.print(e.getMessage());
                printEndSingleTableBox(output);
                printEndTable(output);
            }
        }

        finishViewer(output);
    }

    protected String getName() {
        return "xpath";
    }

    private void printQuery(NodeSet result, ServletOutputStream output, String query, String path)
            throws IOException {

        String resultTitle = "Result for Query " + query;
        printStartTable(resultTitle, output);

        if (!result.hasMoreNodes()) {
            printStartSingleTableBox(output);
            output.print("No result for query " + query + "");
            printEndSingleTableBox(output);
        }

        while (result.hasMoreNodes()) {
            Object node = result.getNextNode();

            printStartSingleTableBox(output);
            if (node instanceof Node) {
                Xml2HtmlWriter.write((Node) node, output);
            } else {
                //	TODO print something, that makes sense
                output.print("Node is no Node");
            }
            printEndSingleTableBox(output);
        }
        printEndTable(output);
    }

    private void printQueryForm(ServletOutputStream output, String query, String path, String[] ns)
            throws IOException {
        String formTitle = "XPath Query for Collection " + path;
        printStartTable(formTitle, output);
        printSearchFromRow("xpath", query, ns, output);
        printEndTable(output);
    }

}
