/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: XUpdateCollectionViewer.java 541515 2007-05-25 02:45:06Z vgritsenko $
 */

package org.apache.xindice.webadmin.viewer.components;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Node;

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.data.NodeSet;
import org.apache.xindice.webadmin.viewer.*;
import org.apache.xindice.webadmin.viewer.HtmlViewerComponent;
import org.apache.xindice.xml.NamespaceMap;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Xindice Html Viewer for collection xupdate searches.
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @version $Revision: 541515 $, $Date: 2007-05-24 22:45:06 -0400 (Thu, 24 May 2007) $
 */
public class XUpdateCollectionViewer extends HtmlViewerComponent implements HtmlCollectionViewer {

    private static final Log log = LogFactory.getLog(XUpdateCollectionViewer.class);

    public void execute(HttpServletRequest req, HttpServletResponse res, Collection col)
            throws ServletException, IOException {

        String path = col.getCanonicalName();
        String title = "XUpdate for " + path;

        ServletOutputStream output = startViewer(title, path, res);

        String xupdate = req.getParameter("xupdate");
        if(xupdate == null) {
            xupdate = "";
        } else if(xupdate.length() > 0) {
            // do xupdate
            try {
                NodeSet result = col.queryCollection("XUpdate", xupdate, new NamespaceMap());
                printResult(result, output);
            } catch (DBException e) {
                log.error(e);
            }
        }

        printStartTable(title, output);
        printStartSingleTableBox(output);
        output.print(XUPDATE_FORM_START);
        output.print(xupdate);
        output.print(XUPDATE_FORM_END);
        printEndSingleTableBox(output);
        printEndTable(output);

        output.println("<br />");
        printStartTable("Examples", output);
        printStartSingleTableBox(output);
        output.println(XUPDATE_EXAMPLES);
        printEndSingleTableBox(output);
        printEndTable(output);

        finishViewer(output);
    }

    private void printResult(NodeSet result, ServletOutputStream output) throws IOException {
        if(result != null && result.hasMoreNodes()) {
            Object node = result.getNextNode();
            String modified;
            if(node instanceof Node) {
                modified = ((Node)node).getFirstChild().getNodeValue();
            } else {
                //	TODO print something, that makes sense
                modified = "Sorry something went wrong!";
            }
            printStartTable("XUpdate Result", output);
            printStartSingleTableBox(output);
            output.println("Modified <b>"+modified+"</b> Nodes!");
            printEndSingleTableBox(output);
            printEndTable(output);
            output.println("<br />");
        }
    }

    private static final String XUPDATE_FORM_START =
        "<form action=\"\" method=\"post\"> \n" +
        "<table class=\"upload-form\"> \n" +
        "<tr><td> \n" +
        "<textarea name=\"xupdate\" cols=\"80\" rows=\"15\">\n" ;

    private static final String XUPDATE_FORM_END =
        "</textarea> \n" +
        "</td></tr><tr><td> \n" +
        "<input type=\"submit\" name=\"submit\" value=\"xupdate\" /> \n" +
        "</td></tr></table></form> \n";

    private static final String XUPDATE_EXAMPLES =
        "<b>Basic XUpdate Insert Command</b>" +
        "<pre>&lt;xupdate:modifications version=\"1.0\" \n" +
        "		xmlns:xupdate=\"http://www.xmldb.org/xupdate\"&gt; \n\n" +
        "   &lt;xupdate:insert-after select=\"/addresses/address[1]\" &gt; \n" +
        "	  &lt;xupdate:element name=\"address\"&gt; \n" +
        "		 &lt;xupdate:attribute name=\"id\"&gt;2&lt;/xupdate:attribute&gt; \n" +
        "		 &lt;fullname&gt;John Smith&lt;/fullname&gt; \n" +
        "		 &lt;born day='2' month='12' year='1974'/&gt; \n" +
        "		 &lt;country&gt;Germany&lt;/country&gt; \n" +
        "	 &lt;/xupdate:element&gt; \n\n" +
        "   &lt;/xupdate:insert-after&gt; \n" +
        "&lt;/xupdate:modifications&gt; \n" +
        " </pre>" +
        "<b>XUpdate Commands</b><ul>" +
        "<li><code>xupdate:insert-before</code> " +
        " - Inserts a new node in document order before the selected node.</li> \n" +
        "<li><code>xupdate:insert-after</code>" +
        " - Inserts a new node in document order after the selected node.</li> \n" +
        "<li><code>xupdate:update</code>" +
        " - Replaces all child nodes of the selected node with the specified nodes.</li> \n" +
        "<li><code>xupdate:append</code>" +
        " - Appends the specified node to the content of the selected node.</li> \n" +
        "<li><code>xupdate:remove</code>" +
        " - Remove the selected node</li> \n" +
        "<li><code>xupdate:rename</code>" +
        " - Renames the selected node</li> \n" +
        "<li><code>xupdate:variable</code>" +
        " - Defines a variable containing a node list that can be reused in later operations.</li> \n" +
        "</ul><b>XUpdate Node Construction</b><ul>" +
        "<li><code>xupdate:element</code>" +
        " - Creates a new element in the document.</li> \n" +
        "<li><code>xupdate:attribute</code>" +
        " - Creates a new attribute node associated with an <code>xupdate:element</code>.</li> \n" +
        "<li><code>xupdate:text</code>" +
        " - Creates a text content node in the document.</li> \n" +
        "<li><code>xupdate:processing-instruction</code>" +
        " - Creates a processing instruction node in the document.</li> \n" +
        "<li><code>xupdate:comment</code>" +
        " - Creates a new comment node in the document.</li> </ul>\n";

    protected String getName() {
        return "xupdate";
    }
}
