/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.xindice.webadmin.webdav;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServletRequest;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;

/**
 * This is a wrapper class that provides additional DAV-specific functionality.
 *
 * @version $Revision: 551440 $, $Date: 2007-06-28 00:26:07 -0400 (Thu, 28 Jun 2007) $
 */
public class DAVRequest {
    public static final int DEPTH_0 = 0;
    public static final int DEPTH_1 = 1;
    public static final int DEPTH_INFINITY = -1;

    private static final Log log = LogFactory.getLog(DAVRequest.class);

    private HttpServletRequest httpRequest;
    private String contextPath;
    private Document requestDoc;
    private boolean documentInit;

    /**
     * Creates new instance of DAVRequest based on incoming HTTP request.
     *
     * @param request HttpServletRequest
     */
    public DAVRequest(HttpServletRequest request) {
        httpRequest = request;
        contextPath = makeContextPath();
    }

    /**
     * @return the path to calling servlet (not null and no ending '/')
     */
    private String makeContextPath() {
        StringBuffer path = new StringBuffer();
        path.append(httpRequest.getContextPath());
        // Is it default servlet? If not, servletPath is added to contextPath path.
        if (httpRequest.getPathInfo() != null) {
            path.append(httpRequest.getServletPath());
        }
        if (path.length() > 0 && path.charAt(path.length() - 1) == '/') {
            path.deleteCharAt(path.length() - 1);
        }

        return path.toString();
    }

    /**
     * Retreives request document if incoming request has any.
     *
     * @return Parsed request document if request body has one or null if
     * request body was empty or does not have a valid XML.
     * @throws IOException ServletInputStream IOException
     */
    public Document getRequestDoc() throws IOException {
        if (!documentInit) {
            documentInit = true;
            requestDoc = parseRequestContent();
        }

        return requestDoc;
    }

    /**
     * Wrapper method, gets resource path.
     *
     * If servlet is mapped as default servlet (with url pattern '/'),
     * resource path is obtained from {@link HttpServletRequest#getServletPath},
     * and {@link HttpServletRequest#getPathInfo} is always null.
     *
     * If servlet is mapped to a path pattern (url pattern like '/foo/*'),
     * resource path is retrieved from {@link HttpServletRequest#getPathInfo}.
     *
     * @return Called servlet path.
     */
    public String getPath() {
        String path = httpRequest.getPathInfo();
        if (path == null) {
            path = httpRequest.getServletPath();
        }
        return path;
    }

    /**
     * Wrapper method, gets HTTP request header.
     * @param name Header name
     * @return Header value
     */
    public String getHeader(String name) {
        return httpRequest.getHeader(name);
    }

    /**
     * Retrieves Overwrite header value from HTTP request. Possible values
     * are "T" or "F" (true of false). By default value is "T".
     *
     * @return False if HTTP request has a header "Overwrite: F", true othewise.
     */
    public boolean getOverwrite() {
        // Parsing overwrite header
        boolean overwrite = true;
        String overwriteHeader = httpRequest.getHeader("Overwrite");
        if (overwriteHeader != null) {
            overwrite = !overwriteHeader.equalsIgnoreCase("F");
        }
        return overwrite;
    }

    /**
     * Retrieves Depth header value from HTTP request. Possible values
     * are "0", "1", and "infinity". By default, value is "infinity".
     *
     * @return 0, 1, -1 for "0", "1", "infinity" header values respectively.
     */
    public int getDepth() {
        String value = httpRequest.getHeader("Depth");
        int depth = -1;
        if ("0".equals(value)) {
            depth = 0;
        } else if ("1".equals(value)) {
            depth = 1;
        } else if (value == null || "infinity".equals(value)) {
            depth = -1;
        }

        return depth;
    }

    /**
     * Retrieves Destination header value from HTTP request. Its value is URL,
     * it can be absolute or relative. This method returns only local path,
     * deleting protocol, server name and context path.
     *
     * @return Local part of the destination path
     */
    public String getDestinationPath() {
        String destinationPath = httpRequest.getHeader("Destination");
        if (destinationPath == null) {
            return null;
        }

        int protocolIndex = destinationPath.indexOf("://");
        if (protocolIndex >= 0) {
            // if the Destination URL contains the protocol, we can safely
            // trim everything upto the first "/" character after "://"
            int firstSeparator = destinationPath.indexOf("/", protocolIndex + 4);
            if (firstSeparator < 0) {
                destinationPath = "/";
            } else {
                destinationPath = destinationPath.substring(firstSeparator);
            }
        } else {
            String hostName = httpRequest.getServerName();
            if ((hostName != null) && (destinationPath.startsWith(hostName))) {
                destinationPath = destinationPath.substring(hostName.length());
            }

            int portIndex = destinationPath.indexOf(":");
            if (portIndex >= 0) {
                destinationPath = destinationPath.substring(portIndex);
            }

            if (destinationPath.startsWith(":")) {
                int firstSeparator = destinationPath.indexOf("/");
                if (firstSeparator < 0) {
                    destinationPath = "/";
                } else {
                    destinationPath = destinationPath.substring(firstSeparator);
                }
            }
        }

        String contextPath = httpRequest.getContextPath();
        if ((contextPath != null) && (destinationPath.startsWith(contextPath))) {
            destinationPath = destinationPath.substring(contextPath.length());
        }

        String pathInfo = httpRequest.getPathInfo();
        if (pathInfo != null) {
            String servletPath = httpRequest.getServletPath();
            if ((servletPath != null) && (destinationPath.startsWith(servletPath))) {
                destinationPath = destinationPath.substring(servletPath.length());
            }
        }

        try {
            destinationPath = URLDecoder.decode(destinationPath, "utf-8");
        } catch (UnsupportedEncodingException e) {
            // won't happen
            log.error(e);
            destinationPath = null;
        }

        return destinationPath;
    }

    /**
     * Wrapper method, gets servlet context path.
     * @return Called servlet context path.
     */
    public String getContextPath() {
        return contextPath;
    }

    private Document parseRequestContent() throws IOException {
        if (httpRequest.getContentLength() != 0) {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            try {
                DocumentBuilder documentBuilder = dbf.newDocumentBuilder();
                return documentBuilder.parse(new InputSource(httpRequest.getInputStream()));
            } catch (IOException e) {
                throw e;
            } catch (Exception e) {
                if (log.isWarnEnabled()) {
                    log.warn("Cannot parse request document", e);
                }
            }
        }

        return null;
    }

    /**
     * Wrapper method, gets servlet input stream.
     * @return Servlet input stream.
     */
    public ServletInputStream getInputStream() throws IOException {
        return httpRequest.getInputStream();
    }

    /**
     * Wrapper method, gets HTTP request parameter.
     *
     * @param name Parameter name
     * @return Parameter value
     */
    public String getParameter(String name) {
        return httpRequest.getParameter(name);
    }

    /**
     * Wrapper method, gets request protocol.
     *
     * @return Request protocol and version.
     */
    public String getProtocol() {
        return httpRequest.getProtocol();
    }
}
