/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Delete.java 541515 2007-05-25 02:45:06Z vgritsenko $
 */

package org.apache.xindice.webadmin.webdav.components;

import java.io.IOException;

import javax.servlet.ServletException;

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.Database;
import org.apache.xindice.webadmin.webdav.WebdavStatus;
import org.apache.xindice.webadmin.webdav.DAVResponse;
import org.apache.xindice.webadmin.webdav.DAVRequest;
import org.apache.xindice.webadmin.Location;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This class implements the Delete command for WebDAV operations on
 * Xindice.<br>
 * <br>
 * DELETE command deletes target collection or resource. For collections,
 * it deletes the entire subtree, so Depth header in request cannot be
 * 0 or 1, but it can be omitted.<br>
 * <br>
 * The default success status code is 204 (No Content)
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @author <a href="mailto:gianugo@apache.org">Gianugo Rabellino</a>
 * @version $Revision: 541515 $, $Date: 2007-05-24 22:45:06 -0400 (Thu, 24 May 2007) $
 */
public class Delete implements DAVComponent {
    private final Log log = LogFactory.getLog(Delete.class);

    public void execute(DAVRequest req, DAVResponse res, Location target) throws ServletException, IOException {
        if (target.isRoot()) {
            res.setStatus(WebdavStatus.SC_FORBIDDEN);
            return;
        }

        Collection col = target.getCollection();
        String name = target.getName();

        if (col == null) {
            res.setStatus(WebdavStatus.SC_NOT_FOUND);
            return;
        }

        if (col instanceof Database) {
            res.setStatus(WebdavStatus.SC_FORBIDDEN);
            return;
        }

        if (req.getDepth() != DAVRequest.DEPTH_INFINITY) {
            res.setStatus(WebdavStatus.SC_BAD_REQUEST);
            return;
        }

        // TODO: set SC_FORBIDDEN for system collections
        if (name == null) { // deleting collection
            try {
                col.getDatabase().dropCollection(col);
            } catch (DBException e) {
                log.error("Failed to delete collection " + col.getCanonicalName(), e);
                throw new ServletException(e);
            }
        } else { // deleting resource
            try {
                col.remove(name);
                if (log.isDebugEnabled()) {
                    log.debug("entry " + name + " deleted");
                }
            } catch (DBException e) {
                if (e.faultCode == FaultCodes.COL_DOCUMENT_NOT_FOUND) {
                    res.setStatus(WebdavStatus.SC_NOT_FOUND);
                    return;
                } else {
                    log.error("Failed to delete resource " + name + " from collection " + col.getCanonicalName(), e);
                    throw new ServletException(e);
                }
            }
        }

        res.setStatus(WebdavStatus.SC_NO_CONTENT);
    }
}
