/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: SymbolTable.java 568786 2007-08-23 00:40:57Z vgritsenko $
 */

package org.apache.xindice.xml;

import org.apache.xindice.util.ReadOnlyException;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * SymbolTable is a class that allows the Xindice Compression system to
 * resolve symbol IDs in compression/decompression passes.
 *
 * @version $Revision: 568786 $, $Date: 2007-08-22 20:40:57 -0400 (Wed, 22 Aug 2007) $
 */
public class SymbolTable implements XMLSerializable {
    static final String SYMBOLS = "symbols";
    static final String SYMBOL = "symbol";
    static final String NAME = "name";
    static final String NSURI = "nsuri";
    static final String ID = "id";

    private transient boolean dirty;
    private transient long lastModified = System.currentTimeMillis();

    private short maxSymbol = -1;
    private final Map symbols = new HashMap(); // String to SymbolInfo
    private final Map names = new HashMap();   // Short to SymbolInfo
    private boolean readOnly;


    public static final class SymbolInfo implements Serializable {
        private final String namespaceURI;
        private final String qname;
        private final short symbol;

        private SymbolInfo(String qname, short symbol) {
            this.namespaceURI = null;
            this.qname = qname;
            this.symbol = symbol;
        }

        private SymbolInfo(String qname, String namespaceURI, short symbol) {
            this.namespaceURI = namespaceURI;
            this.qname = qname;
            this.symbol = symbol;
        }

        public String getNamespaceURI() {
            return namespaceURI;
        }

        public String getQName() {
            return qname;
        }

        public short getSymbolID() {
            return symbol;
        }
    }

    //
    // Instance Methods
    //

    public SymbolTable() {
    }

    public SymbolTable(Element symbols) {
        streamFromXML(symbols);
    }

    public SymbolTable(Element symbols, boolean readOnly) {
        this(symbols);
        this.readOnly = readOnly;
    }

    /**
     * @return true if symbol table has been modified
     */
    public final boolean isDirty() {
        return dirty;
    }

    /**
     * Set dirty flag to true and update last modified
     * time stamp.
     */
    private void markDirty() {
        this.dirty = true;
        this.lastModified = System.currentTimeMillis();
    }

    /**
     * Reset dirty flag to false. Should be invoked
     * once symbol table is successfully saved.
     */
    public final void resetDirty() {
        this.dirty = false;
    }

    public final long getLastModified() {
        return lastModified;
    }

    /**
     * @return true if symbol table is read only
     */
    public final boolean isReadOnly() {
        return readOnly;
    }

    //
    // Lookup by symbol id
    //

    public final SymbolInfo getSymbolInfo(short symbol) {
        return (SymbolInfo) names.get(new Short(symbol));
    }

    public final String getNamespaceURI(short symbol) {
        SymbolInfo info = getSymbolInfo(symbol);
        return info != null ? info.namespaceURI : null;
    }

    public final String getName(short symbol) {
        SymbolInfo info = getSymbolInfo(symbol);
        return info != null ? info.qname : null;
    }

    //
    // Lookup by qname and namespace
    //

    public final short getSymbol(String qname) {
        try {
            return getSymbol(qname, false);
        } catch (ReadOnlyException e) {
            // Won't happen
            throw new IllegalStateException();
        }
    }

    public final short getSymbol(String qname, String namespaceURI) {
        try {
            return getSymbol(qname, namespaceURI, false);
        } catch (ReadOnlyException e) {
            // Won't happen
            throw new IllegalStateException();
        }
    }

    public final short getSymbol(String qname, boolean create) throws ReadOnlyException {
        SymbolInfo info = (SymbolInfo) symbols.get(qname);
        if (info != null) {
            return info.symbol;
        }

        if (create) {
            if (readOnly) {
                throw new ReadOnlyException();
            }

            synchronized (symbols) {
                short id = ++maxSymbol;
                info = new SymbolInfo(qname, id);
                symbols.put(qname, info);
                names.put(new Short(id), info);
                markDirty();
                return id;
            }
        }

        return -1;
    }

    public final short getSymbol(String qname, String namespaceURI, boolean create) throws ReadOnlyException {
        String lookupName = getLookupName(qname, namespaceURI);

        SymbolInfo info = (SymbolInfo) symbols.get(lookupName);
        if (info != null) {
            return info.symbol;
        }

        if (create) {
            if (readOnly) {
                throw new ReadOnlyException();
            }

            synchronized (symbols) {
                short id = ++maxSymbol;
                info = new SymbolInfo(qname, namespaceURI, id);
                symbols.put(lookupName, info);
                names.put(new Short(id), info);
                markDirty();
                return id;
            }
        }

        return -1;
    }

    //
    // Lookup normalized symbol
    //

    public final short getNormalizedSymbol(String localName, String namespaceURI) {
        try {
            return getNormalizedSymbol(localName, namespaceURI, false);
        } catch (ReadOnlyException e) {
            // Won't happen
            throw new IllegalStateException();
        }
    }

    /**
     * Lookup normalized symbol by element (or attribute) local name and namespace URI.
     *
     * @param localName element (attribute) local name
     * @param namespaceURI element (attribute) namespace URI
     * @param create when true, creates symbol if it is missing
     * @return symbol id or -1
     * @throws ReadOnlyException if 'create' option specified and the symbol is missing
     */
    public final short getNormalizedSymbol(String localName, String namespaceURI, boolean create) throws ReadOnlyException {
        String normalizedQName = getNormalizedQName(localName, namespaceURI);
        return getSymbol(normalizedQName, namespaceURI, create);
    }

    /**
     * Lookup normalized symbol by lookup string and (optional) namespace map. Lookup
     * string can take one of the following forms:
     * <ul>
     * <li> [<namespaceURI>]<nsPrefix>:<localName>, where nsPrefix can be empty
     * <li> <nsPrefix>:<localName>, any regular qName
     * </ul>
     *
     * Namespace map is used to determine namespaceURI for passed in lookup string and
     * normalize namespace prefix.
     *
     * @param lookup lookup string
     * @param nsMap namespace map
     * @param create when true, creates symbol if it is missing
     * @return symbol id or -1
     * @throws ReadOnlyException if 'create' option specified and the symbol is missing
     */
    public final short getNormalizedSymbol(String lookup, NamespaceMap nsMap, boolean create) throws ReadOnlyException {
        // Parse [<namespaceURI>]<nsPrefix>:<localName> with optional nsPrefix
        if (lookup.startsWith("[")) {
            int idx = lookup.indexOf(']');
            String nsURI = lookup.substring(1, idx);
            int cidx = lookup.indexOf(':', idx + 1);
            String name = cidx != -1 ? lookup.substring(cidx + 1) : lookup.substring(idx + 1);

            return getSymbol(getNormalizedQName(name, nsURI), nsURI, create);
        }

        // Parse <nsPrefix>:<localName> with passed in nsMap
        int idx = lookup.indexOf(':');
        if (idx != -1) {
            String pfx = lookup.substring(0, idx);
            String nsURI = (String) nsMap.get(pfx);
            if (nsURI != null) {
                String name = lookup.substring(idx + 1);
                return getSymbol(getNormalizedQName(name, nsURI), nsURI, create);
            }
        }

        return getSymbol(lookup, create);
    }

    //
    // XMLSerializable
    //

    public final Element streamToXML(Document doc) throws DOMException {
        Element root = doc.createElement(SYMBOLS);

        synchronized (symbols) {
            Iterator i = symbols.values().iterator();
            while (i.hasNext()) {
                SymbolInfo info = (SymbolInfo) i.next();

                Element e = doc.createElement(SYMBOL);
                e.setAttribute(ID, Short.toString(info.symbol));
                e.setAttribute(NAME, info.qname);
                if (info.namespaceURI != null && info.namespaceURI.length() > 0) {
                    e.setAttribute(NSURI, info.namespaceURI);
                }
                root.appendChild(e);
            }
        }
        
        return root;
    }

    public final void streamFromXML(Element element) throws DOMException {
        symbols.clear();
        names.clear();
        maxSymbol = -1;

        NodeList list = element.getElementsByTagName(SYMBOL);
        int size = list.getLength();
        for (int i = 0; i < size; i++) {
            Element elem = (Element) list.item(i);

            String qname = elem.getAttribute(NAME);
            String namespaceURI = elem.getAttribute(NSURI);
            if (namespaceURI != null && namespaceURI.length() == 0) {
                namespaceURI = null;
            }

            short id = Short.parseShort(elem.getAttribute(ID));
            if (id > maxSymbol) {
                maxSymbol = id;
            }

            SymbolInfo info = new SymbolInfo(qname, namespaceURI, id);
            if (namespaceURI != null) {
                String lookupName = getLookupName(qname, namespaceURI);
                symbols.put(lookupName, info);
            } else {
                symbols.put(qname, info);
            }

            names.put(new Short(id), info);
        }
    }

    //
    // Implementation methods
    //

    private String getNormalizedQName(String localName, String namespaceURI) {
        return "ns" + namespaceURI.hashCode() + ':' + localName;
    }

    private String getLookupName(String qname, String namespaceURI) {
        return '[' + namespaceURI + ']' + qname;
    }
}
