/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DOMCompressor.java 568786 2007-08-23 00:40:57Z vgritsenko $
 */

package org.apache.xindice.xml.dom;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.util.ReadOnlyException;
import org.apache.xindice.xml.Signatures;
import org.apache.xindice.xml.SymbolTable;
import org.apache.xindice.xml.XMLCompressedOutput;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * DOMCompressor is an OutputStream extension that provides functions for
 * writing DOM types to a Xindice Compressed XML Stream.
 *
 * @version $Revision: 568786 $, $Date: 2007-08-22 20:40:57 -0400 (Wed, 22 Aug 2007) $
 */
public final class DOMCompressor extends XMLCompressedOutput {

    private static final Log log = LogFactory.getLog(DOMCompressor.class);


    public DOMCompressor(OutputStream os, SymbolTable st) {
        super(os, st);
    }

    /**
     * writeNode writes a Node to the compressed output stream.  This method
     * is recursive and will write all children of the specific Node.
     *
     * @param node the node to write
     * @throws IOException if the write to underlying stream has failed
     * @throws DOMException if used symbol table is read only and some of
     *         node's content can not be serialized.
     */
    public void writeNode(Node node) throws IOException {
        // Check if it's node of ours
        if (node instanceof NodeImpl) {
            NodeImpl impl = (NodeImpl) node;
            Document doc = node.getOwnerDocument();

            SymbolTable docSymbols = null;
            if (doc instanceof DocumentImpl) {
                docSymbols = ((DocumentImpl) doc).getSymbols();
            }

            // If document is compressed with exact same symbol table
            // and is not dirty, use its compressed bytes.
            if (!impl.dirty && impl.data != null && docSymbols == st) {
                write(impl.data, impl.pos, impl.len);
                flush();
                return;
            }
        }

        // Have to do it manually
        switch (node.getNodeType()) {

            case Node.ELEMENT_NODE:
                {
                    String nsURI = node.getNamespaceURI();
                    short symbolID;
                    if (nsURI != null) {
                        try {
                            symbolID = st.getSymbol(node.getNodeName(), nsURI, true);
                        } catch (ReadOnlyException e) {
                            throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
                                                   "Element can not be serialized (read only symbol table). " +
                                                   "Name: '" + node.getNodeName() + "', namespace: " + nsURI);
                        }
                    } else {
                        try {
                            symbolID = st.getSymbol(node.getNodeName(), true);
                        } catch (ReadOnlyException e) {
                            throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
                                                   "Element can not be serialized (read only symbol table). " +
                                                   "Name: '" + node.getNodeName() + "'");
                        }
                    }

                    byte signature = Signatures.NODE_ELEM;
                    int attrLen = node.getAttributes().getLength();
                    int attrSize = getSizeType(attrLen);
                    if (attrLen > 0) {
                        signature |= Signatures.ELEM_ATTRS;
                    }
                    if (node.hasChildNodes()) {
                        signature |= Signatures.ELEM_CHILDREN;
                    }
                    signature |= (byte) (attrSize);
                    byte[] children = buildChildren(node);
                    int valLen = children.length;
                    int sizeType = getSizeType(valLen + 11);
                    signature |= (byte) (sizeType << 0x2);
                    writeByte(signature);
                    valLen += (getSizeSize(sizeType) + 3);
                    writeSize(sizeType, valLen);
                    writeShort(symbolID);
                    write(children);
                    break;
                }

            case Node.ATTRIBUTE_NODE:
                {
                    Attr attr = (Attr) node;
                    String nsURI = attr.getNamespaceURI();
                    short symbolID;
                    if (nsURI != null) {
                        try {
                            symbolID = st.getSymbol(attr.getName(), nsURI, true);
                        } catch (ReadOnlyException e) {
                            throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
                                                   "Attribute can not be serialized (read only symbol table). " +
                                                   "Name: '" + attr.getName() + "', namespace: " + nsURI);
                        }
                    } else {
                        try {
                            symbolID = st.getSymbol(attr.getName(), true);
                        } catch (ReadOnlyException e) {
                            throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
                                                   "Attribute can not be serialized (read only symbol table). " +
                                                   "Name: '" + attr.getName() + "'");
                        }
                    }
                    writeShort(symbolID);

                    byte[] b = attr.getValue().getBytes("UTF-8");
                    writeShort((short) b.length);
                    write(b);
                    break;
                }

            case Node.TEXT_NODE:
                {
                    byte[] b = node.getNodeValue().getBytes("UTF-8");
                    int valLen = b.length;
                    int sizeType = getSizeType(valLen + 5);
                    byte signature = Signatures.NODE_TEXT;
                    signature |= (byte) (sizeType << 0x2);
                    writeByte(signature);
                    valLen += (getSizeSize(sizeType) + 1);
                    writeSize(sizeType, valLen);
                    write(b);
                    break;
                }

            case Node.CDATA_SECTION_NODE:
                {
                    byte[] b = node.getNodeValue().getBytes("UTF-8");
                    int valLen = b.length;
                    writeByte(Signatures.NODE_DECL | Signatures.DECL_CDATA);
                    valLen += 5;
                    writeInt(valLen);
                    write(b);
                    break;
                }

            case Node.ENTITY_REFERENCE_NODE:
                {
                    String value = node.getNodeName();
                    byte signature = /* Signatures.NODE_TEXT | */ Signatures.TEXT_ENTITY;
                    short symbol = 0;
                    int encoding = 0;
                    if (value.equals("&amp;")) {
                        signature |= (byte) (Signatures.ENT_AMP);
                    } else if (value.equals("&lt;")) {
                        signature |= (byte) (Signatures.ENT_LT);
                    } else if (value.equals("&gt;")) {
                        signature |= (byte) (Signatures.ENT_GT);
                    } else if (value.equals("&quot;")) {
                        signature |= (byte) (Signatures.ENT_QUOT);
                    } else if (value.equals("&apos;")) {
                        signature |= (byte) (Signatures.ENT_APOS);
                    } else if (value.startsWith("&#x")) {
                        encoding = 1;
                        signature |= (byte) (Signatures.ENT_UNICODE);
                        // Convert the Unicode to a short
                        // TODO symbol =
                    } else {
                        encoding = 2;
                        try {
                            symbol = st.getSymbol(value, true);
                        } catch (ReadOnlyException e) {
                            throw new DOMException(DOMException.NOT_SUPPORTED_ERR,
                                                   "Entity reference can not be serialized (read only symbol table). " +
                                                   "Value: '" + value + "'");
                        }
                        signature |= (byte) (Signatures.ENT_DEFINED);
                    }
                    writeByte(signature);
                    if (encoding > 0) {
                        writeShort(symbol);
                    }
                    break;
                }

            case Node.ENTITY_NODE:
                {
                    break;
                }

            case Node.PROCESSING_INSTRUCTION_NODE:
                {
                    String value = node.getNodeName() + " " + node.getNodeValue();
                    byte[] b = value.getBytes("UTF-8");
                    int valLen = b.length;
                    writeByte(Signatures.NODE_PROC);
                    valLen += 5;
                    writeInt(valLen);
                    write(b);
                    break;
                }

            case Node.COMMENT_NODE:
                {
                    byte[] b = node.getNodeValue().getBytes("UTF-8");
                    int valLen = b.length;
                    writeByte(Signatures.NODE_DECL | Signatures.DECL_COMMENT);
                    valLen += 5;
                    writeInt(valLen);
                    write(b);
                    break;
                }

            case Node.DOCUMENT_NODE:
                {
                    byte[] children = buildChildren(node);
                    writeInt(children.length);
                    write(children);
                    break;
                }

            case Node.DOCUMENT_TYPE_NODE:
                {
                    // Errr?
                    break;
                }

            case Node.DOCUMENT_FRAGMENT_NODE:
                {
                    byte[] children = buildChildren(node);
                    writeInt(children.length);
                    write(children);
                    break;
                }

            case Node.NOTATION_NODE:
                {
                    // byte signature = (byte)(Signatures.Decl << 0x6);
                    break;
                }

            default:
                if (log.isWarnEnabled()) {
                    log.warn("invalid node type : " + node.getNodeType());
                }
        }
        flush();
    }

    /**
     * buildChildren builds a byte array containing the child definitions
     * for the specified Node.  Because Xindice's compressed system is a
     * depth-first system that requires the length of nested data sets in
     * order to allow quick traversal, this process must build the byte
     * arrays in memory before including them in their parents.
     * <br>
     * If a Node has already been graphed and has not been modified, it's
     * byte array portion is simply copied into the stream without
     * requiring a regraph. This allows for quickly streaming out a modified
     * DOM tree, but will  take some time in streaming a brand new DOM tree.
     *
     * @param node The Node whose children to build
     * @return The child byte array
     * @throws IOException If the build failed
     */
    private byte[] buildChildren(Node node) throws IOException {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        DOMCompressor out = new DOMCompressor(bos, st);
        if (node.getNodeType() == Node.ELEMENT_NODE) {
            NamedNodeMap attrs = node.getAttributes();
            int len = attrs.getLength();
            out.writeSize(len);
            for (int i = 0; i < len; i++) {
                out.writeNode(attrs.item(i));
            }
        }

        NodeList children = node.getChildNodes();
        int len = children.getLength();
        for (int i = 0; i < len; i++) {
            out.writeNode(children.item(i));
        }

        return bos.toByteArray();
    }

    /**
     * Compress is a convenience method that compresses a Node into a byte
     * array with a single call.
     *
     * @param node The Node to compress
     * @param symbols The Symbol Table to use
     * @return The resulting byte array
     */
    public static byte[] compress(Node node, SymbolTable symbols) {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        DOMCompressor xco = new DOMCompressor(bos, symbols);
        node.normalize();
        try {
            xco.writeNode(node);
        } catch (IOException e) {
            // ByteArrayOutputStream does not throw IOException, ignore
        }
        return bos.toByteArray();
    }
}


