/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DocumentImpl.java 571948 2007-09-02 10:51:37Z vgritsenko $
 */

package org.apache.xindice.xml.dom;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.xml.NodeSource;
import org.apache.xindice.xml.SymbolTable;
import org.apache.xindice.xml.dom.traversal.TreeWalkerImpl;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.DOMConfiguration;
import org.w3c.dom.DOMException;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;
import org.w3c.dom.UserDataHandler;
import org.w3c.dom.traversal.DocumentTraversal;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;
import org.w3c.dom.traversal.TreeWalker;

import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * DocumentImpl
 *
 * @version $Revision: 571948 $, $Date: 2007-09-02 06:51:37 -0400 (Sun, 02 Sep 2007) $
 */
public final class DocumentImpl extends ContainerNodeImpl
                                implements CompressedDocument, DBDocument, DocumentTraversal {

    private static final Log log = LogFactory.getLog(DocumentImpl.class);

    private DocumentType docType;
    private String version;
    private String xmlEncoding;
    private String inputEncoding;
    private boolean standalone;
    private boolean strictErrorChecking;
    private SymbolTable symbols;
    private boolean readOnly;
    private DOMConfiguration domConfig = new DOMConfigurationImpl();

    /**
     * Create empty dirty document.
     */
    public DocumentImpl() {
        super(null, true);
    }

    /**
     * Create document from the compressed data.
     *
     * @param data compressed document data
     * @param symbols symbol table used to compress a document
     */
    public DocumentImpl(byte[] data, SymbolTable symbols) {
        super(null, data, 0, data.length);
        this.symbols = symbols;
    }

    /**
     * Create document from the compressed data.
     *
     * @param data compressed document data
     * @param symbols symbol table used to compress a document
     * @param source identifies document origin
     */
    public DocumentImpl(byte[] data, SymbolTable symbols, NodeSource source) {
        this(data, symbols);
        this.source = source;
    }

    /**
     * Create document from the compressed data with specified read only state.
     *
     * @param data compressed document data
     * @param symbols symbol table used to compress a document
     * @param source identifies document origin
     * @param readOnly if true, document will be marked read only.
     */
    public DocumentImpl(byte[] data, SymbolTable symbols, NodeSource source, boolean readOnly) {
        this(data, symbols, source);
        this.readOnly = readOnly;
    }

    /**
     * Create a compressed document out of another document.
     *
     * @param doc document to copy
     */
    public DocumentImpl(Document doc) {
        super(null, true);

        boolean compress = true;
        if (doc instanceof CompressedDocument) {
            CompressedDocument c = (CompressedDocument) doc;
            symbols = c.getSymbols();
            if (!c.isDirty()) {
                data = c.getDataBytes();
                pos = c.getDataPos();
                len = c.getDataLen();
                compress = false;
            }
        }

        if (compress) {
            if (symbols == null) {
                symbols = new SymbolTable();
            }
            data = DOMCompressor.compress(doc, symbols);
            pos = 0;
            len = data.length;
        }

        if (doc instanceof DBDocument) {
            source = ((DBDocument) doc).getSource();
        }
    }

    public boolean isReadOnly() {
        return readOnly;
    }

    protected void checkLoaded() {
        if (loaded) {
            return;
        }

        loaded = true;
        try {
            if (data != null) {
                loadChildren(symbols);
            }
        } catch (IOException e) {
            if (log.isWarnEnabled()) {
                log.warn("ignored exception", e);
            }
        }
    }

    public boolean isCaching() {
        String cache = DBDocument.CACHE;

        int size = childNodes.getLength();
        for (int i = 0; i < size; i++) {
            Node n = childNodes.item(i);
            if (n.getNodeType() == Node.PROCESSING_INSTRUCTION_NODE && n.getNodeName().equals(CACHE_CONTROL)) {
                cache = n.getNodeValue().trim();
                break;
            }
        }

        return (cache != null && cache.equals(CACHE));
    }

    public void setCaching(boolean caching) {
        int size = childNodes.getLength();
        for (int i = 0; i < size; i++) {
            Node n = childNodes.item(i);
            if (n.getNodeType() == Node.PROCESSING_INSTRUCTION_NODE && n.getNodeName().equals(CACHE_CONTROL)) {
                n.setNodeValue(caching ? CACHE : NOCACHE);
                return;
            }
        }

        ProcessingInstruction pi = createProcessingInstruction(CACHE_CONTROL, caching ? CACHE : NOCACHE);
        insertBefore(pi, getDocumentElement());
    }

    public SymbolTable getSymbols() {
        return symbols;
    }

    public void setSymbols(SymbolTable symbols) {
        this.symbols = symbols;
    }

    public void expandSource() {
        ElementImpl e = (ElementImpl) getDocumentElement();
        if (e != null) {
            e.expandSource();
        }
    }

    public Node getNodeAtPos(int pos) {
        return null; // TODO: This
    }

    public short getNodeType() {
        return Node.DOCUMENT_NODE;
    }

    public String getNodeName() {
        return "#document";
    }

    /**
     * The Document Type Declaration (see <code>DocumentType</code>) associated
     * with  this document. For HTML documents as well as XML documents without
     * a document type declaration this returns <code>null</code>. The DOM Level
     *  1 does not support editing the Document Type Declaration, therefore
     * <code>docType</code> cannot be altered in any way.
     */
    public DocumentType getDoctype() {
        return docType;
    }

    public void setDoctype(DocumentType docType) {
        this.docType = docType;
    }

    /**
     * The <code>DOMImplementation</code> object that handles this document. A
     * DOM application may use objects from multiple  implementations.
     */
    public DOMImplementation getImplementation() {
        return DOMImplementationImpl.getInstance();
    }

    /**
     * Creates a <code>Text</code> node given the specified string.
     * @param data The data for the node.
     * @return The new <code>Text</code> object.
     */
    public Text createTextNode(String data) {
        return new TextImpl(this, data);
    }

    /**
     * This is a convenience method that allows direct access to the child
     * node that is the root element of  the document. For HTML documents, this
     * is the element with the tagName "HTML".
     */
    public Element getDocumentElement() {
        checkLoaded();

        Iterator i = childNodes.iterator();
        while (i.hasNext()) {
            Node node = (Node) i.next();
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                return (Element) node;
            }
        }

        return null;
    }

    /**
     * Creates a <code>CDATASection</code> node whose value  is the specified
     * string.
     * @param data The data for the <code>CDATASection</code> contents.
     * @return The new <code>CDATASection</code> object.
     * @exception DOMException
     *   NOT_SUPPORTED_ERR: Raised if this document is an HTML document.
     */
    public CDATASection createCDATASection(String data) throws DOMException {
        return new CDATASectionImpl(this, data);
    }

    /**
     * Creates an element of the type specified. Note that the instance returned
     * implements the Element interface, so attributes can be specified
     * directly  on the returned object.
     * @param tagName The name of the element type to instantiate. For XML, this
     *   is case-sensitive. For HTML, the  <code>tagName</code> parameter may
     *   be provided in any case,  but it must be mapped to the canonical
     *   uppercase form by  the DOM implementation.
     * @return A new <code>Element</code> object.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified name contains an
     *   invalid character.
     */
    public Element createElement(String tagName) throws DOMException {
        return new ElementImpl(this, tagName);
    }

    /**
     * Creates an empty <code>DocumentFragment</code> object.
     * @return A new <code>DocumentFragment</code>.
     */
    public DocumentFragment createDocumentFragment() {
        return new DocumentFragmentImpl(this);
    }

    /**
     * Creates an <code>Attr</code> of the given name. Note that the
     * <code>Attr</code> instance can then be set on an <code>Element</code>
     * using the <code>setAttribute</code> method.
     * @param name The name of the attribute.
     * @return A new <code>Attr</code> object.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified name contains an
     *   invalid character.
     */
    public Attr createAttribute(String name) throws DOMException {
        return new AttrImpl(this, name);
    }

    /**
     * Creates a <code>Comment</code> node given the specified string.
     * @param data The data for the node.
     * @return The new <code>Comment</code> object.
     */
    public Comment createComment(String data) {
        return new CommentImpl(this, data);
    }

    /**
     * Creates a <code>ProcessingInstruction</code> node given the specified
     * name and data strings.
     * @param target The target part of the processing instruction.
     * @param data The data for the node.
     * @return The new <code>ProcessingInstruction</code> object.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if an invalid character is specified.
     *   <br>NOT_SUPPORTED_ERR: Raised if this document is an HTML document.
     */
    public ProcessingInstruction createProcessingInstruction(String target, String data) throws DOMException {
        return new ProcessingInstructionImpl(this, target, data);
    }

    /**
     * Creates an EntityReference object.
     * @param name The name of the entity to reference.
     * @return The new <code>EntityReference</code> object.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified name contains an
     *   invalid character.
     *   <br>NOT_SUPPORTED_ERR: Raised if this document is an HTML document.
     */
    public EntityReference createEntityReference(String name) throws DOMException {
        return new EntityReferenceImpl(this, name);
    }

    private void importNamespaces(Node source, Node target) {
        if (target.getNodeType() == Node.ELEMENT_NODE) {
            // Retrieve Namespace definitions in scope
            Set set = new HashSet();
            Node n = source;
            Element elem = (Element) target;
            while (n != null) {
                NamedNodeMap nm = n.getAttributes();
                for (int i = 0; i < nm.getLength(); i++) {
                    Attr a = (Attr) nm.item(i);
                    String name = a.getNodeName();
                    if ((name.startsWith("xmlns:") || name.equals("xmlns")) && !set.contains(name)) {
                        set.add(name);
                        elem.setAttribute(name, a.getValue());
                    }
                }
                n = n.getParentNode();
                if (n == null || n.getNodeType() == DOCUMENT_NODE || n.getNodeType() == DOCUMENT_FRAGMENT_NODE) {
                    n = null;
                }
            }
        }
    }

    public Node importNode(Node importedNode, boolean deep) {
        return importNode(importedNode, deep, true, true);
    }

    private Node importNode(Node importedNode, boolean deep, boolean importNamespaces, boolean invokeHandler) {
        try {
            // If we're a Xindice Compressed DOM Node, and share the same symbol table,
            // then we're golden
            if (importedNode instanceof NodeImpl) {
                NodeImpl impl = (NodeImpl) importedNode;
                DocumentImpl docImpl = (DocumentImpl) impl.getOwnerDocument();
                if (docImpl.getSymbols() != null && (docImpl.getSymbols() == symbols)) {
                    NodeImpl clone = (NodeImpl) impl.cloneNode(deep, false);
                    clone.setParentNode(this);

                    if (importNamespaces) {
                        importNamespaces(importedNode, clone);
                    }
                    if (invokeHandler) {
                        invokeHandlers(UserDataHandler.NODE_IMPORTED, importedNode, clone);
                    }

                    return clone;
                }
            }

            // Crap, we have to do a full graph copy
            Node result = null;
            switch (importedNode.getNodeType()) {
                case Node.ATTRIBUTE_NODE:
                    result = createAttribute(importedNode.getNodeName());
                    break;

                case Node.CDATA_SECTION_NODE:
                    result = createCDATASection(importedNode.getNodeValue());
                    break;

                case Node.COMMENT_NODE:
                    result = createComment(importedNode.getNodeValue());
                    break;

                case Node.ELEMENT_NODE:
                    Element selem = (Element) importedNode;
                    Element elem = createElement(selem.getTagName());
                    NamedNodeMap attrs = selem.getAttributes();
                    int size = attrs.getLength();
                    for (int i = 0; i < size; i++) {
                        Attr a = (Attr) attrs.item(i);
                        Attr ai = createAttribute(a.getName());
                        ai.setValue(a.getValue());
                        elem.setAttributeNode(ai);
                    }
                    result = elem;
                    if (importNamespaces) {
                        importNamespaces(importedNode, result);
                    }
                    break;

                case Node.ENTITY_REFERENCE_NODE:
                    result = createEntityReference(importedNode.getNodeValue());
                    break;

                case Node.PROCESSING_INSTRUCTION_NODE:
                    result = createProcessingInstruction(importedNode.getNodeName(), importedNode.getNodeValue());
                    break;

                case Node.TEXT_NODE:
                    result = createTextNode(importedNode.getNodeValue());
                    break;

                default:
                    // FIXME Huh?
            }

            if (deep && result != null) {
                NodeList list = importedNode.getChildNodes();
                int size = list.getLength();
                for (int i = 0; i < size; i++) {
                    Node n = list.item(i);
                    result.appendChild(importNode(n, deep, false, invokeHandler));
                }

                if (invokeHandler) {
                    invokeHandlers(UserDataHandler.NODE_IMPORTED, importedNode, result);
                }
            }

            return result;
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("ignored exception", e);
            }
            return null;
        }
    }

    public Element createElementNS(String namespaceURI, String qualifiedName) {
        return new ElementImpl(this, qualifiedName);
    }

    public Attr createAttributeNS(String namespaceURI, String qualifiedName) {
        return new AttrImpl(this, qualifiedName);
    }

    public NodeIterator createNodeIterator(Node root, int whatToShow, NodeFilter filter, boolean entityReferenceExpansion) throws DOMException {
        return new TreeWalkerImpl(root, whatToShow, filter, entityReferenceExpansion);
    }

    public TreeWalker createTreeWalker(Node root, int whatToShow, NodeFilter filter, boolean entityReferenceExpansion) throws DOMException {
        return new TreeWalkerImpl(root, whatToShow, filter, entityReferenceExpansion);
    }

    //
    // DOM Level 3 Implementation
    //

    /**
     * @since DOM Level 3
     */
    public Node adoptNode(Node src) {
        // If we're a Xindice DOM Node and share the same symbol table
        // or the adopted node has no symbol table, then we're golden
        if (src instanceof NodeImpl) {
            NodeImpl impl = (NodeImpl) src;
            DocumentImpl docImpl = (DocumentImpl) impl.getOwnerDocument();
            if (docImpl.getSymbols() == null || docImpl.getSymbols() == symbols) {
                impl.getParentNode().removeChild(impl);
                impl.setParentNode(this);
                invokeHandlers(UserDataHandler.NODE_ADOPTED, src, null);
                return impl;
            }
        }

        Node node = importNode(src, true, true, false);
        invokeHandlers(UserDataHandler.NODE_ADOPTED, src, node);

        return node;
    }

    /**
     * @since DOM Level 3
     */
    public boolean getStrictErrorChecking() {
        checkLoaded();
        return strictErrorChecking;
    }

    /**
     * @since DOM Level 3
     */
    public void setStrictErrorChecking(boolean strictErrorChecking) {
        checkReadOnly();
        checkLoaded();
        this.strictErrorChecking = strictErrorChecking;
    }

    /**
     * @since DOM Level 3
     */
    public String getDocumentURI() {
        return null;
    }

    /**
     * @since DOM Level 3
     */
    public void setDocumentURI(String documentURI) {
    }

    /**
     * @since DOM Level 3
     */
    public String getInputEncoding() {
        checkLoaded();
        return inputEncoding;
    }

    public void setInputEncoding(String inputEncoding) {
        checkReadOnly();
        checkLoaded();
        this.inputEncoding = inputEncoding;
    }

    /**
     * @since DOM Level 3
     */
    public String getXmlEncoding() {
        checkLoaded();
        return xmlEncoding;
    }

    public void setXmlEncoding(String xmlEncoding) {
        checkReadOnly();
        checkLoaded();
        this.xmlEncoding = xmlEncoding;
    }

    /**
     * @since DOM Level 3
     */
    public boolean getXmlStandalone() {
        checkLoaded();
        return standalone;
    }

    /**
     * @since DOM Level 3
     */
    public void setXmlStandalone(boolean standalone) throws DOMException {
        checkReadOnly();
        checkLoaded();
        this.standalone = standalone;
    }

    /**
     * @since DOM Level 3
     */
    public String getXmlVersion() {
        checkLoaded();
        return version;
    }

    /**
     * @since DOM Level 3
     */
    public void setXmlVersion(String version) throws DOMException {
        checkReadOnly();
        checkLoaded();
        this.version = version;
    }

    /**
     * @since DOM Level 3
     */
    public DOMConfiguration getDomConfig() {
        return domConfig;
    }

    /**
     * @since DOM Level 3
     */
    public void normalizeDocument() {
    }

    /**
     * Rename an existing node of type ELEMENT_NODE or ATTRIBUTE_NODE.
     * <p/>
     * When possible this simply changes the name of the given node, otherwise
     * this creates a new node with the specified name and replaces the
     * existing node with the new node as described below.
     * <p/>
     * If simply changing the name of the given node is not possible, the
     * following operations are performed: a new node is created, any
     * registered event listener is registered on the new node, any user data
     * attached to the old node is removed from that node, the old node is
     * removed from its parent if it has one, the children are moved to the
     * new node, if the renamed node is an Element its attributes are moved to
     * the new node, the new node is inserted at the position the old node used
     * to have in its parent's child nodes list if it has one, the user data
     * that was attached to the old node is attached to the new node.
     * <p/>
     * When the node being renamed is an Element only the specified attributes
     * are moved, default attributes originated from the DTD are updated
     * according to the new element name. In addition, the implementation may
     * update default attributes from other schemas. Applications should use
     * Document.normalizeDocument() to guarantee these attributes are
     * up-to-date.
     * <p/>
     * When the node being renamed is an Attr that is attached to an Element,
     * the node is first removed from the Element attributes map. Then, once
     * renamed, either by modifying the existing node or creating a new one as
     * described above, it is put back.
     * <p/>
     * In addition, <ul>
     * <li>a user data event NODE_RENAMED is fired,
     * <li>when the implementation supports the feature "MutationNameEvents",
     * each mutation operation involved in this method fires the appropriate
     * event, and in the end the event {http://www.w3.org/2001/xml-events,
     * DOMElementNameChanged} or {http://www.w3.org/2001/xml-events,
     * DOMAttributeNameChanged} is fired.
     * </ul>
     * @param n The node to rename.
     * @param namespaceURI The new namespace URI.
     * @param qualifiedName The new qualified name.
     * @return The renamed node. This is either the specified node or the new
     * node that was created to replace the specified node.
     * @throws DOMException NOT_SUPPORTED_ERR: Raised when the type of the
     * specified node is neither ELEMENT_NODE nor ATTRIBUTE_NODE, or if the
     * implementation does not support the renaming of the document element.
     * @throws DOMException INVALID_CHARACTER_ERR: Raised if the new qualified
     * name is not an XML name according to the XML version in use specified
     * in the Document.xmlVersion attribute.
     * @throws DOMException WRONG_DOCUMENT_ERR: Raised when the specified node
     * was created from a different document than this document.
     * @throws DOMException NAMESPACE_ERR: Raised if the qualifiedName is a
     * malformed qualified name, if the qualifiedName has a prefix and the
     * namespaceURI is null, or if the qualifiedName has a prefix that is
     * "xml" and the namespaceURI is different from
     * "http://www.w3.org/XML/1998/namespace" [XML Namespaces]. Also raised,
     * when the node being renamed is an attribute, if the qualifiedName, or
     * its prefix, is "xmlns" and the namespaceURI is different from
     * "http://www.w3.org/2000/xmlns/".
     * @since DOM Level 3
     */
    public Node renameNode(Node n, String namespaceURI, String qualifiedName) throws DOMException {
        if (!(n instanceof NodeImpl) || n.getOwnerDocument() != this) {
            throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, "Node was created from a different document");
        }

        int idx = qualifiedName.indexOf(':');
        String prefix = idx >= 0 ? qualifiedName.substring(0, idx) : null;

        if (prefix != null && ("".equals(prefix) || namespaceURI == null || "".equals(namespaceURI))) {
            throw new DOMException(DOMException.NAMESPACE_ERR, "Malformed qualified name '" + qualifiedName + "'.");
        }

        return ((NodeImpl) n).renameNode(namespaceURI, qualifiedName, prefix);
    }

    /**
     * @since DOM Level 3
     */
    public String getTextContent() {
        return null;
    }
}
