/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: NamedNodeMapImpl.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.xml.dom;

import org.w3c.dom.DOMException;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import java.util.HashMap;
import java.util.Iterator;

/**
 * NamedNodeMapImpl
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public final class NamedNodeMapImpl extends NodeListImpl implements NamedNodeMap {
    private HashMap map = new HashMap();

    public NamedNodeMapImpl(NodeImpl owner) {
        super(owner);
    }

    /**
     * Retrieves a node specified by name.
     * @param name Name of a node to retrieve.
     * @return A <code>Node</code> (of any type) with the specified name, or
     *   <code>null</code> if the specified name did not identify any node in
     *   the map.
     */
    public Node getNamedItem(String name) {
        return (Node) map.get(name);
    }

    /**
     * Adds a node using its <code>nodeName</code> attribute.
     * <br>As the <code>nodeName</code> attribute is used to derive the name
     * which the node must be stored under, multiple nodes of certain types
     * (those that have a "special" string value) cannot be stored as the names
     * would clash. This is seen as preferable to allowing nodes to be aliased.
     * @param arg A node to store in a named node map. The node will later be
     *   accessible using the value of the <code>nodeName</code> attribute of
     *   the node. If a node with that name is already present in the map, it
     *   is replaced by the new one.
     * @return If the new <code>Node</code> replaces an existing node with the
     *   same name  the previously existing <code>Node</code> is returned,
     *   otherwise <code>null</code> is returned.
     * @exception DOMException
     *   WRONG_DOCUMENT_ERR: Raised if <code>arg</code> was created from a
     *   different document than the one that created the
     *   <code>NamedNodeMap</code>.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this
     *   <code>NamedNodeMap</code> is readonly.
     *   <br>INUSE_ATTRIBUTE_ERR: Raised if <code>arg</code> is an
     *   <code>Attr</code> that is already an attribute of another
     *   <code>Element</code> object. The DOM user must explicitly clone
     *   <code>Attr</code> nodes to re-use them in other elements.
     */
    public Node setNamedItem(Node arg) throws DOMException {
        String name = arg.getNodeName();
        Node node = (Node) map.get(name);
        if (node != null) {
            int idx = indexOf(node);
            set(idx, arg);
        } else {
            add(arg);
        }
        map.put(arg.getNodeName(), arg);
        return node;
    }

    /**
     * Removes a node specified by name. If the removed node is an
     * <code>Attr</code> with a default value it is immediately replaced.
     * @param name The name of a node to remove.
     * @return The node removed from the map or <code>null</code> if no node
     *   with such a name exists.
     * @exception DOMException
     *   NOT_FOUND_ERR: Raised if there is no node named <code>name</code> in
     *   the map.
     */
    public Node removeNamedItem(String name) throws DOMException {
        Node node = (Node) map.get(name);
        map.remove(name);
        remove(node);
        return node;
    }

    public Node getNamedItemNS(String namespaceURI, String localName) {
        Node node;
        for (Iterator iter = iterator(); iter.hasNext(); ) {
            node = (Node) iter.next();
            String ns = node.getNamespaceURI();
            String ln = node.getLocalName();
            if (ns != null
                    && ln != null
                    && ns.equals(namespaceURI)
                    && ln.equals(localName)) {
                return node;
            }
        }
        return null;
    }

    public Node setNamedItemNS(Node arg) throws DOMException {
        return setNamedItem(arg);
    }

    /**
     * Removes an attribute by local name and namespace URI. If the
     * removed attribute has a default value it is immediately
     * replaced. The replacing attribute has the same namespace URI
     * and local name, as well as the original prefix.  as the
     * original prefix. If the named attribute is not present, this
     * method has no effect (per clarification in DOM2 errata).
     *
     * <br> HTML-only DOM implementations do not need to implement
     * this method.
     *
     * @param namespaceURI  The  namespace URI of the attribute to remove.
     * @param localName  The  local name of the attribute to remove.
     **/
    public Node removeNamedItemNS(String namespaceURI, String localName) throws DOMException {
        Node node;
        for (Iterator iter = iterator(); iter.hasNext(); ) {
            node = (Node) iter.next();
            if (localName.equals(node.getLocalName())) {
                if ((namespaceURI == null && node.getNamespaceURI() == null) ||
                        (namespaceURI != null && namespaceURI.equals(node.getNamespaceURI()))) {
                    return removeNamedItem(node.getNodeName());
                }
            }
        }
        return null;
    }
}

