/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: SetContentHandler.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.xml.sax;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xmldb.api.modules.XMLResource;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Simple ContentHandler that just converts the SAX event stream into a text
 * representation of the document and stores it in the associated resource.
 *
 * <small>
 * The only place this class currently used is in
 * {@link org.apache.xindice.client.xmldb.resources.XMLResourceImpl#setContentAsSAX}.
 * Instead of this class, null-transform using TrAX API can be used.
 * </small>
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public class SetContentHandler extends DefaultHandler {

    private static final Log log = LogFactory.getLog(SetContentHandler.class);

    protected XMLResource resource = null;
    protected StringBuffer newContent = null;
    protected Map namespaces = null;

    public SetContentHandler(XMLResource resource) {
        this.resource = resource;
        namespaces = new HashMap();
    }

    /**
     * Receive notification of the beginning of the document.
     *
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#startDocument
     */
    public void startDocument()
            throws SAXException {
        newContent = new StringBuffer();
        newContent.append("<?xml version=\"1.0\"?>");
    }


    /**
     * Receive notification of the end of the document.
     *
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#endDocument
     */
    public void endDocument()
            throws SAXException {
        try {
            resource.setContent(newContent.toString());
        } catch (Exception e) {
            log.warn("Exception in endDocument", e);
            throw new SAXException(e);
        }
    }


    /**
     * Receive notification of the start of a Namespace mapping.
     *
     * @param prefix The Namespace prefix being declared.
     * @param uri The Namespace URI mapped to the prefix.
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#startPrefixMapping
     */
    public void startPrefixMapping(String prefix, String uri)
            throws SAXException {
        namespaces.put(prefix, uri);
    }


    /**
     * Receive notification of the end of a Namespace mapping.
     *
     * @param prefix The Namespace prefix being declared.
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#endPrefixMapping
     */
    public void endPrefixMapping(String prefix)
            throws SAXException {
        namespaces.remove(prefix);
    }


    private String getLocalName(String qn) {
        if (qn.indexOf(':') != -1) {
            return qn.substring(qn.indexOf(':') + 1);
        } else {
            return qn;
        }
    }

    private String getQNameAtt(String uri, String localName) throws SAXException {

        String prefix = null;

        if ("".equals(uri)) {
            return localName;
        }

        /* Look for prefix */
        Iterator prefixes = namespaces.keySet().iterator();
        while (prefixes.hasNext()) {
            String key = (String) prefixes.next();
            if ((!("".equals(key))) && namespaces.get(key).equals(uri)) {
                prefix = key;
                break;
            }
        }

        if (prefix == null) {
            throw new SAXException("No declared prefix for namespace '" +
                                   uri + "'.");
        }

        return prefix + ":" + localName;
    }


    private String getQNameElement(String uri, String localName) throws SAXException {
        String prefix = null;

        if ("".equals(uri)) {
            if (namespaces.get("") != null) {
                throw new SAXException("default namespace is declared here!");
            } else {

                return localName;
            }
        }

        /* Look for prefix */
        Iterator prefixes = namespaces.keySet().iterator();
        while (prefixes.hasNext()) {
            String key = (String) prefixes.next();
            if (namespaces.get(key).equals(uri)) {
                prefix = key;
                break;
            }
        }

        if (prefix == null) {
            throw new SAXException("No declared prefix for namespace '"
                                   + uri + "'.");
        }

        return ("".equals(prefix) ? localName : prefix + ":" + localName);
    }


    /**
     * Receive notification of the start of an element.
     *
     * @param attributes The specified or defaulted attributes.
     * @param uri Description of Parameter
     * @param localName Description of Parameter
     * @param qName Description of Parameter
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#startElement
     */
    public void startElement(String uri, String localName,
                             String qName, Attributes attributes)
            throws SAXException {

        newContent.append("<");

        /* Make up a correct qName if necessary */
        if ("".equals(qName)) {
            newContent.append(getQNameElement(uri, localName));
        } else {
            newContent.append(qName);
        }

        for (int i = 0; i < attributes.getLength(); i++) {
            String qn = attributes.getQName(i);

            // Make up a correct qName if necessary
            if ("".equals(qn)) {
                qn = getQNameAtt(attributes.getURI(i), attributes.getLocalName(i));
            }
            newContent.append(" ");
            newContent.append(qn);
            newContent.append("=");
            newContent.append("\"");
            newContent.append(attributes.getValue(i));
            newContent.append("\"");

            // Avoid duplicate namespace declarations
            if (qn.equals("xmlns")) {
                namespaces.remove("");
            }

            if (qn.startsWith("xmlns:")) {
                String ln = getLocalName(qn);
                namespaces.remove(ln);
            }
        }

        Iterator iter = namespaces.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            newContent.append(" xmlns");
            if (key.length() > 0) {
                newContent.append(":");
                newContent.append(key);
            }
            newContent.append("=");
            newContent.append("\"");
            newContent.append(namespaces.get(key));
            newContent.append("\"");
            namespaces.remove(key);
        }

        newContent.append(">");
    }


    /**
     * Receive notification of the end of an element.
     *
     * @param uri Description of Parameter
     * @param localName Description of Parameter
     * @param qName Description of Parameter
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#endElement
     */
    public void endElement(String uri, String localName, String qName)
            throws SAXException {
        newContent.append("</");

        if ("".equals(qName)) {
            qName = getQNameElement(uri, localName);
        }
        newContent.append(qName);
        newContent.append(">");
    }


    /**
     * Receive notification of character data inside an element.
     *
     * @param ch The characters.
     * @param start The start position in the character array.
     * @param length The number of characters to use from the
     * character array.
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#characters
     */
    public void characters(char ch[], int start, int length)
            throws SAXException {
        int i = 0;
        while (i < length) {
            char c = ch[start + i];
            switch (c) {
                case '&':
                    newContent.append("&amp;");
                    break;
                case '<':
                    newContent.append("&lt;");
                    break;
                case '>':
                    newContent.append("&gt;");
                    break;
                case '"':
                    newContent.append("&quot;");
                    break;
                case '\'':
                    newContent.append("&apos;");
                    break;
                default:
                    // If we're outside 7 bit ascii encode as a character ref.
                    // Not sure what the proper behavior here should be.
                    if ((int) c > 127) {
                        newContent.append("&#" + (int) c + ";");
                    } else {
                        newContent.append(c);
                    }
            }

            i++;
        }
    }


    /**
     * Receive notification of ignorable whitespace in element content.
     *
     * @param ch The whitespace characters.
     * @param start The start position in the character array.
     * @param length The number of characters to use from the
     * character array.
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#ignorableWhitespace
     */
    public void ignorableWhitespace(char ch[], int start, int length)
            throws SAXException {
        int i = 0;
        while (i < length) {
            newContent.append(ch[start + i]);
            i++;
        }
    }


    /**
     * Receive notification of a processing instruction.
     *
     * @param target The processing instruction target.
     * @param data The processing instruction data, or null if
     * none is supplied.
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#processingInstruction
     */
    public void processingInstruction(String target, String data)
            throws SAXException {
        newContent.append("<?");
        newContent.append(target);
        newContent.append(" ");

        if (data != null) {
            newContent.append(data);
        }

        newContent.append("?>");
    }


    /**
     * Receive notification of a skipped entity.
     *
     * @param name The name of the skipped entity.
     * @exception SAXException Description of Exception
     * @see org.xml.sax.ContentHandler#processingInstruction
     */
    public void skippedEntity(String name)
            throws SAXException {
        // no op
    }
}
