/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.xindice.webadmin;

import org.apache.xindice.webadmin.webdav.components.Propfind;
import org.apache.xindice.webadmin.webdav.components.props.CreationDate;
import org.apache.xindice.webadmin.webdav.components.props.LastModified;
import org.apache.xindice.webadmin.webdav.WebdavStatus;
import org.apache.xindice.xml.dom.DOMParser;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.core.Collection;
import org.apache.xindice.core.data.Entry;
import org.custommonkey.xmlunit.XMLAssert;
import org.xml.sax.SAXException;

public class PropfindTest extends MethodSetup {
    private CreationDate creationDate = new CreationDate();
    private LastModified lastModified = new LastModified();
    private final String DOC_NAME1 = "testDoc1";
    private final String DOC_NAME2 = "testDoc2";
    private Collection child;

    public void setUp() throws Exception {
        super.setUp();
        String doc = "<test>Test document</test>";
        collection.insertDocument(DOC_NAME1, DOMParser.toDocument(doc));
        child = collection.createCollection("child", new Configuration(
                DOMParser.toDocument(
                        "<collection compressed='true' name='" + "child" + "' inline-metadata='true'>" +
                        "<filer class='org.apache.xindice.core.filer.BTreeFiler' />" +
                        "</collection>"), false
        ));

        child.insertDocument(DOC_NAME2, DOMParser.toDocument(doc));
        method = new Propfind();
    }

    public void testListNullCollection() throws Exception {
        // source collection was not found
        method.execute(request, response, new Location("/unknown"));
        assertEquals(WebdavStatus.SC_NOT_FOUND, httpResponse.getStatus());
    }

    public void testListNullResource() throws Exception {
        String path = collection.getCanonicalDocumentName("unknown");
        httpRequest.setPathInfo(path);
        httpRequest.setInput("<?xml version='1.0' encoding='utf-8' ?>" +
                             "<propfind xmlns='DAV:'><allprop/></propfind>");
        method.execute(request, response, new Location(path));
        assertEquals(WebdavStatus.SC_NOT_FOUND, httpResponse.getStatus());
    }

    public void testListEmptyRequestBody() throws Exception {
        httpRequest.setPathInfo(collection.getCanonicalName());
        httpRequest.setInput("<?xml version='1.0' encoding='utf-8' ?>" +
                             "<propfind xmlns='DAV:'><allprop/></propfind>");
        method.execute(request, response, new Location(collection.getCanonicalName()));

        Entry entry1 = collection.getEntry(DOC_NAME1);
        Entry entry2 = child.getEntry(DOC_NAME2);
        assertEquals(WebdavStatus.SC_MULTI_STATUS, httpResponse.getStatus());
        String expected =
                "<?xml version='1.0' encoding='utf-8' ?>" +
                "<multistatus xmlns='DAV:'>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype><collection/></resourcetype>" +
                                "<getlastmodified>" + lastModified.getProperty(collection).getValue() + "</getlastmodified>" +
                                "<displayname>" + collection.getName() + "</displayname>" +
                                "<creationdate>" + creationDate.getProperty(collection).getValue() + "</creationdate>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalDocumentName(DOC_NAME1) + "</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<getcontenttype>text/xml</getcontenttype>" +
                                "<getlastmodified>" + lastModified.getProperty(entry1).getValue() + "</getlastmodified>" +
                                "<displayname>" + DOC_NAME1 + "</displayname>" +
                                "<creationdate>" + creationDate.getProperty(entry1).getValue() + "</creationdate>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + child.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype><collection/></resourcetype>" +
                                "<getlastmodified>" + lastModified.getProperty(child).getValue() + "</getlastmodified>" +
                                "<displayname>" + child.getName() + "</displayname>" +
                                "<creationdate>" + creationDate.getProperty(child).getValue() + "</creationdate>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + child.getCanonicalDocumentName(DOC_NAME2) + "</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<getcontenttype>text/xml</getcontenttype>" +
                                "<getlastmodified>" + lastModified.getProperty(entry2).getValue() + "</getlastmodified>" +
                                "<displayname>" + DOC_NAME2 + "</displayname>" +
                                "<creationdate>" + creationDate.getProperty(entry2).getValue() + "</creationdate>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                "</multistatus>";

        assertXMLEqual(expected, httpResponse.getResponse());
    }

    public void testListAllProp() throws Exception {
        // request does not have a body, should be treated as allprop
        httpRequest.addHeader("Depth", "0");
        httpRequest.setPathInfo(collection.getCanonicalName());
        method.execute(request, response, new Location(collection.getCanonicalName()));

        assertEquals(WebdavStatus.SC_MULTI_STATUS, httpResponse.getStatus());
        String expected =
                "<?xml version='1.0' encoding='utf-8' ?>" +
                "<multistatus xmlns='DAV:'>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype><collection/></resourcetype>" +
                                "<getlastmodified>" + lastModified.getProperty(collection).getValue() + "</getlastmodified>" +
                                "<displayname>" + collection.getName() + "</displayname>" +
                                "<creationdate>" + creationDate.getProperty(collection).getValue() + "</creationdate>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                "</multistatus>";

        assertXMLEqual(expected, httpResponse.getResponse());
    }

    public void testListNamedProperties() throws Exception {
        httpRequest.addHeader("Depth", "infinity");
        httpRequest.setPathInfo(collection.getCanonicalName());
        httpRequest.setInput("<?xml version='1.0' encoding='utf-8' ?>" +
                             "<propfind xmlns='DAV:'>" +
                             "<prop>" +
                             "<displayname/><resourcetype/><unknown/><getcontenttype/>" +
                             "</prop>" +
                             "</propfind>");
        method.execute(request, response, new Location(collection.getCanonicalName()));

        assertEquals(WebdavStatus.SC_MULTI_STATUS, httpResponse.getStatus());
        String expected =
                "<?xml version='1.0' encoding='utf-8' ?>" +
                "<multistatus xmlns='DAV:'>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype><collection/></resourcetype>" +
                                "<displayname>" + collection.getName() + "</displayname>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                        "<propstat>" +
                            "<prop>" +
                                "<unknown/>" +
                                "<getcontenttype/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 404 Not Found</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalDocumentName(DOC_NAME1) + "</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<getcontenttype>text/xml</getcontenttype>" +
                                "<displayname>" + DOC_NAME1 + "</displayname>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                        "<propstat>" +
                            "<prop>" +
                                "<unknown/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 404 Not Found</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + child.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype><collection/></resourcetype>" +
                                "<displayname>" + child.getName() + "</displayname>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                        "<propstat>" +
                            "<prop>" +
                                "<unknown/>" +
                                "<getcontenttype/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 404 Not Found</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + child.getCanonicalDocumentName(DOC_NAME2) + "</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<getcontenttype>text/xml</getcontenttype>" +
                                "<displayname>" + DOC_NAME2 + "</displayname>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                        "<propstat>" +
                            "<prop>" +
                                "<unknown/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 404 Not Found</status>" +
                        "</propstat>" +
                    "</response>\n" +
                "</multistatus>";

        assertXMLEqual(expected, httpResponse.getResponse());
    }

    public void testListPropertyNames() throws Exception {
        httpRequest.addHeader("Depth", "1");
        httpRequest.setPathInfo(collection.getCanonicalName());
        httpRequest.setInput("<?xml version='1.0' encoding='utf-8' ?>" +
                             "<propfind xmlns='DAV:'><propname/></propfind>");
        method.execute(request, response, new Location(collection.getCanonicalName()));

        assertEquals(WebdavStatus.SC_MULTI_STATUS, httpResponse.getStatus());
        String expected =
                "<?xml version='1.0' encoding='utf-8' ?>" +
                "<multistatus xmlns='DAV:'>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<displayname/>" +
                                "<getlastmodified/>" +
                                "<creationdate/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + collection.getCanonicalDocumentName(DOC_NAME1) + "</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<displayname/>" +
                                "<getcontenttype/>" +
                                "<getlastmodified/>" +
                                "<creationdate/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                    "<response>" +
                        "<href>" + child.getCanonicalName() + "/</href>" +
                        "<propstat>" +
                            "<prop>" +
                                "<resourcetype/>" +
                                "<displayname/>" +
                                "<getlastmodified/>" +
                                "<creationdate/>" +
                            "</prop>" +
                            "<status>HTTP/1.1 200 OK</status>" +
                        "</propstat>" +
                    "</response>\n" +
                "</multistatus>";

        assertXMLEqual(expected, httpResponse.getResponse());
    }

    protected void assertXMLEqual(String control, String result) throws Exception {
        try {
            XMLAssert.assertXMLEqual(control, result);
        } catch (SAXException e) {
            fail(e.getLocalizedMessage());
        }
    }
}
